# coding: utf-8
# 2021/8/1 @ tongshiwei

import json
import os.path
from typing import List, Tuple
from EduNLP.constant import MODEL_DIR
from ..Vector import T2V, get_pretrained_t2v as get_t2v_pretrained_model
from ..Vector import get_pretrained_model_info, get_all_pretrained_models
from longling import path_append
from EduData import get_data
from ..Tokenizer import Tokenizer, get_tokenizer
from EduNLP.Pretrain import ElmoTokenizer, BertTokenizer, DisenQTokenizer, QuesNetTokenizer, Question
from EduNLP import logger

__all__ = ["I2V", "D2V", "W2V", "Elmo", "Bert", "DisenQ", "QuesNet", "get_pretrained_i2v"]


class I2V(object):
    """
    It just a api, so you shouldn't use it directly. \
    If you want to get vector from item, you can use other model like D2V and W2V.

    Parameters
    -----------
    tokenizer: str
        the name of tokenizer. eg. bert, pure_text, ...
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        - True: use pretrained t2v model
        - False: use your own t2v model
    model_dir: str
        local directionary for saving online pretrained models, work only when `pretrained_t2v=True`
    kwargs:
        the parameters passed to t2v

    Examples
    --------
    >>> item = {"如图来自古希腊数学家希波克拉底所研究的几何图形．此图由三个半圆构成，三个半圆的直径分别为直角三角形$ABC$的斜边$BC$, \
    ... 直角边$AB$, $AC$.$\\bigtriangleup ABC$的三边所围成的区域记为$I$,黑色部分记为$II$, 其余部分记为$III$.在整个图形中随机取一点，\
    ... 此点取自$I,II,III$的概率分别记为$p_1,p_2,p_3$,则$\\SIFChoice$$\\FigureID{1}$"}
    >>> model_dir = "examples/test_model/d2v"
    >>> url, model_name, *args = get_pretrained_model_info('d2v_test_256')
    >>> (); path = get_data(url, model_dir); () # doctest: +ELLIPSIS
    (...)
    >>> path = path_append(path, os.path.basename(path) + '.bin', to_str=True)
    >>> i2v = D2V("pure_text", "d2v", filepath=path, pretrained_t2v=False)
    >>> i2v(item)
    ([array([ ...dtype=float32)], None)

    Returns
    -------
    i2v model: I2V
    """

    def __init__(self, tokenizer, t2v, *args, tokenizer_kwargs: dict = None,
                 pretrained_t2v=False, model_dir=MODEL_DIR, **kwargs):
        if pretrained_t2v:
            logger.info("Use pretrained t2v model %s" % t2v)
            self.t2v = get_t2v_pretrained_model(t2v, model_dir)
        else:
            self.t2v = T2V(t2v, *args, **kwargs)
        if tokenizer == 'bert':
            self.tokenizer = BertTokenizer.from_pretrained(
                **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        elif tokenizer == 'quesnet':
            self.tokenizer = QuesNetTokenizer.from_pretrained(
                **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        elif tokenizer == 'elmo':
            self.tokenizer = ElmoTokenizer.from_pretrained(
                **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        elif tokenizer == 'disenq':
            self.tokenizer = DisenQTokenizer.from_pretrained(
                **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        else:
            self.tokenizer: Tokenizer = get_tokenizer(tokenizer,
                                                      **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        self.params = {
            "tokenizer": tokenizer,
            "tokenizer_kwargs": tokenizer_kwargs,
            "t2v": t2v,
            "args": args,
            "kwargs": kwargs,
            "pretrained_t2v": pretrained_t2v
        }

    def __call__(self, items, *args, **kwargs):
        """transfer item to vector"""
        return self.infer_vector(items, *args, **kwargs)

    def tokenize(self, items, *args, key=lambda x: x, **kwargs) -> list:
        # """tokenize item"""
        return self.tokenizer(items, *args, key=key, **kwargs)

    def infer_vector(self, items, key=lambda x: x, **kwargs) -> tuple:
        raise NotImplementedError

    def infer_item_vector(self, tokens, *args, **kwargs) -> ...:
        return self.infer_vector(tokens, *args, **kwargs)[0]

    def infer_token_vector(self, tokens, *args, **kwargs) -> ...:
        return self.infer_vector(tokens, *args, **kwargs)[1]

    def save(self, config_path):
        with open(config_path, "w", encoding="utf-8") as wf:
            json.dump(self.params, wf, ensure_ascii=False, indent=2)

    @classmethod
    def load(cls, config_path, *args, **kwargs):
        with open(config_path, encoding="utf-8") as f:
            params: dict = json.load(f)
            tokenizer = params.pop("tokenizer")
            t2v = params.pop("t2v")
            _args = params.pop("args")
            _kwargs = params.pop("kwargs")
            params.update(_kwargs)
            return cls(tokenizer, t2v, *_args, **params)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        raise NotImplementedError

    @property
    def vector_size(self):
        return self.t2v.vector_size


class D2V(I2V):
    """
    The model aims to transfer item to vector directly.

    Bases
    -------
    I2V

    Parameters
    -----------
    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        True: use pretrained t2v model
        False: use your own t2v model
    kwargs:
        the parameters passed to t2v

    Examples
    ---------
    >>> item = {"如图来自古希腊数学家希波克拉底所研究的几何图形．此图由三个半圆构成，三个半圆的直径分别为直角三角形$ABC$的斜边$BC$, \
    ... 直角边$AB$, $AC$.$\\bigtriangleup ABC$的三边所围成的区域记为$I$,黑色部分记为$II$, 其余部分记为$III$.在整个图形中随机取一点，\
    ... 此点取自$I,II,III$的概率分别记为$p_1,p_2,p_3$,则$\\SIFChoice$$\\FigureID{1}$"}
    >>> model_dir = "examples/test_model/d2v"
    >>> url, model_name, *args = get_pretrained_model_info('d2v_test_256')
    >>> (); path = get_data(url, model_dir); () # doctest: +ELLIPSIS
    (...)
    >>> path = path_append(path, os.path.basename(path) + '.bin', to_str=True)
    >>> i2v = D2V("pure_text","d2v",filepath=path, pretrained_t2v = False)
    >>> i2v(item)
    ([array([ ...dtype=float32)], None)

    Returns
    -------
    i2v model: I2V
    """

    def infer_vector(self, items, tokenize=True, key=lambda x: x, *args,
                     **kwargs) -> tuple:
        """
        It is a function to switch item to vector. And before using the function, it is necessary to load model.

        Parameters
        -----------
        items:str
            the text of question
        tokenize: bool
            True: tokenize the item
        key: function
            determine how to get the text of each item
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v

        Returns
        --------
        vector:list
        """
        tokens = self.tokenize(items, key=key) if tokenize is True else items
        tokens = [token for token in tokens]
        return self.t2v(tokens, *args, **kwargs), None

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        return cls("pure_text", name, pretrained_t2v=True, model_dir=model_dir)


class W2V(I2V):
    """
    The model aims to transfer tokens to vector.

    Bases
    --------
    I2V

    Parameters
    -----------
    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        True: use pretrained t2v model
        False: use your own t2v model
    kwargs:
        the parameters passed to t2v

    Examples
    ---------
    >>> (); i2v = get_pretrained_i2v("w2v_test_256", "examples/test_model/w2v"); () # doctest: +SKIP
    (...)
    >>> item_vector, token_vector = i2v(["有学者认为：‘学习’，必须适应实际"]) # doctest: +SKIP
    >>> item_vector # doctest: +SKIP
    [array([...], dtype=float32)]

    Returns
    --------
    i2v model: W2V

    """

    def infer_vector(self, items, tokenize=True, key=lambda x: x, *args,
                     **kwargs) -> tuple:
        '''
        It is a function to switch item to vector. And before using the function, it is necessary to load model.

        Parameters
        -----------
        items:str
            the text of question
        tokenize:bool
            True: tokenize the item
        key: function
            determine how to get the text of each item
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v

        Returns
        --------
        vector:list
        '''
        tokens = self.tokenize(items) if tokenize is True else items
        tokens = [token for token in tokens]
        return self.t2v(tokens, *args, **kwargs), self.t2v.infer_tokens(tokens, *args, **kwargs)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        return cls("pure_text", name, pretrained_t2v=True, model_dir=model_dir)


class Elmo(I2V):
    """The model aims to transfer item and tokens to vector with Elmo.

    Bases
    -------
    I2V

    Parameters
    -----------

    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        True: use pretrained t2v model
        False: use your own t2v model
    kwargs:
        the parameters passed to t2v

    Returns
    -------
    i2v model: Elmo
    """

    def infer_vector(self, items: Tuple[List[str], List[dict], str, dict],
                     *args, key=lambda x: x, **kwargs) -> tuple:
        """It is a function to switch item to vector. And before using the function, it is necessary to load model.

        Parameters
        -----------
        items : str or dict or list
            the item of question, or question list
        return_tensors: str
            tensor type used in tokenizer
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v

        Returns
        --------
        vector: list
        """
        is_batch = isinstance(items, list)
        items = items if is_batch else [items]
        inputs = self.tokenize(items, key=key)
        return self.t2v.infer_vector(inputs, *args, **kwargs), self.t2v.infer_tokens(inputs, *args, **kwargs)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        model_path = path_append(model_dir, get_pretrained_model_info(name)[0].split('/')[-1], to_str=True)
        for i in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar"]:
            model_path = model_path.replace(i, "")
        logger.info("model_path: %s" % model_path)
        tokenizer_kwargs = {"tokenizer_config_dir": model_path}
        return cls("elmo", name, pretrained_t2v=True, model_dir=model_dir,
                   tokenizer_kwargs=tokenizer_kwargs)


class Bert(I2V):
    """
    The model aims to transfer item and tokens to vector with Bert.

    Bases
    -------
    I2V

    Parameters
    -----------
    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        True: use pretrained t2v model
        False: use your own t2v model
    kwargs:
        the parameters passed to t2v

    Returns
    -------
    i2v model: Bert
    """

    def infer_vector(self, items: Tuple[List[str], List[dict], str, dict],
                     *args, key=lambda x: x, return_tensors='pt', **kwargs) -> tuple:
        """
        It is a function to switch item to vector. And before using the function, it is nesseary to load model.

        Parameters
        -----------
        items : str or dict or list
            the item of question, or question list
        return_tensors: str
            tensor type used in tokenizer
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v

        Returns
        --------
        vector:list
        """
        inputs = self.tokenize(items, key=key, return_tensors=return_tensors)
        return self.t2v.infer_vector(inputs, *args, **kwargs), self.t2v.infer_tokens(inputs, *args, **kwargs)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        model_path = path_append(model_dir, get_pretrained_model_info(name)[0].split('/')[-1], to_str=True)
        for i in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar"]:
            model_path = model_path.replace(i, "")
        logger.info("model_path: %s" % model_path)
        tokenizer_kwargs = {"tokenizer_config_dir": model_path}
        return cls("bert", name, pretrained_t2v=True, model_dir=model_dir,
                   tokenizer_kwargs=tokenizer_kwargs)


class DisenQ(I2V):
    """
    The model aims to transfer item and tokens to vector with DisenQ.
    Bases
    -------
    I2V
    Parameters
    -----------
    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
        True: use pretrained t2v model
        False: use your own t2v model
    kwargs:
        the parameters passed to t2v
    Returns
    -------
    i2v model: DisenQ
    """

    def infer_vector(self, items: Tuple[List[str], List[dict], str, dict],
                     *args, key=lambda x: x, vector_type=None, **kwargs) -> tuple:
        """
        It is a function to switch item to vector. And before using the function, it is nesseary to load model.
        Parameters
        -----------
        items : str or dict or list
            the item of question, or question list
        key: function
            determine how to get the text of each item
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v
        Returns
        --------
        vector:list
        """
        is_batch = isinstance(items, list)
        items = items if is_batch else [items]
        inputs = self.tokenize(items, key=key, **kwargs)
        i_vec = self.t2v.infer_vector(inputs, vector_type=vector_type, **kwargs)
        t_vec = self.t2v.infer_tokens(inputs, **kwargs)
        return i_vec, t_vec

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, **kwargs):
        model_path = path_append(model_dir, get_pretrained_model_info(name)[0].split('/')[-1], to_str=True)
        for i in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar"]:
            model_path = model_path.replace(i, "")
        logger.info("model_dir: %s" % model_path)

        tokenizer_kwargs = {
            "tokenizer_config_dir": model_path,
        }
        return cls("disenq", name, pretrained_t2v=True, model_dir=model_dir,
                   tokenizer_kwargs=tokenizer_kwargs, **kwargs)


class QuesNet(I2V):
    """
    The model aims to transfer item and tokens to vector with quesnet.
    Bases
    -------
    I2V
    """

    def infer_vector(self, items: Tuple[List[str], List[dict], str, dict],
                     *args, key=lambda x: x, meta=['know_name'], **kwargs):
        """ It is a function to switch item to vector. And before using the function, it is nesseary to load model.
        Parameters
        ----------
        items : str or dict or list
            the item of question, or question list
        tokenize : bool, optional
            True: tokenize the item
        key : function, optional
            determine how to get the text of each item, by default lambdax: x
        meta : list, optional
            meta information, by default ['know_name']
        args:
            the parameters passed to t2v
        kwargs:
            the parameters passed to t2v
        Returns
        -------
        token embeddings
        question embedding
        """
        encodes = self.tokenize(items, key=key, meta=meta, *args, **kwargs)
        return self.t2v.infer_vector(encodes), self.t2v.infer_tokens(encodes)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        model_path = path_append(model_dir, get_pretrained_model_info(name)[0].split('/')[-1], to_str=True)
        for i in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar"]:
            model_path = model_path.replace(i, "")
        logger.info("model_path: %s" % model_path)
        tokenizer_kwargs = {
            "tokenizer_config_dir": model_path}
        return cls("quesnet", name, pretrained_t2v=True, model_dir=model_dir,
                   tokenizer_kwargs=tokenizer_kwargs)


MODEL_MAP = {
    "w2v": W2V,
    "d2v": D2V,
    "bert": Bert,
    "disenq": DisenQ,
    "quesnet": QuesNet,
    "elmo": Elmo
}


def get_pretrained_i2v(name, model_dir=MODEL_DIR):
    """
    It is a good idea if you want to switch item to vector earily.

    Parameters
    -----------
    name: str
        the name of item2vector model
        e.g.:
        d2v_math_300
        w2v_math_300
        elmo_math_2048
        bert_math_768
        bert_taledu_768
        disenq_math_256
        quesnet_math_512
    model_dir:str
        the path of model, default: MODEL_DIR = '~/.EduNLP/model'

    Returns
    --------
    i2v model: I2V

    Examples
    ---------
    >>> item = {"如图来自古希腊数学家希波克拉底所研究的几何图形．此图由三个半圆构成，三个半圆的直径分别为直角三角形$ABC$的斜边$BC$, \
    ... 直角边$AB$, $AC$.$\\bigtriangleup ABC$的三边所围成的区域记为$I$,黑色部分记为$II$, 其余部分记为$III$.在整个图形中随机取一点，\
    ... 此点取自$I,II,III$的概率分别记为$p_1,p_2,p_3$,则$\\SIFChoice$$\\FigureID{1}$"}
    >>> (); i2v = get_pretrained_i2v("d2v_test_256", "examples/test_model/d2v"); () # doctest: +SKIP
    (...)
    >>> print(i2v(item)) # doctest: +SKIP
    ([array([ ...dtype=float32)], None)
    """
    pretrained_models = get_all_pretrained_models()
    if name not in pretrained_models:
        raise KeyError(
            "Unknown model name %s, use one of the provided models: %s" % (name, ", ".join(pretrained_models))
        )
    _, t2v = get_pretrained_model_info(name)
    _class, *params = MODEL_MAP[t2v], name
    return _class.from_pretrained(*params, model_dir=model_dir)



import os
import json
from transformers.file_utils import TensorType
from transformers import AutoTokenizer
from typing import List, Optional, Union, Tuple
from ..SIF import EDU_SPYMBOLS
from ..Tokenizer import get_tokenizer


class TokenizerForHuggingface(object):
    """
    Parameterss
    ----------
    pretrained_model:
        used pretrained model
    add_specials:
        Whether to add tokens like [FIGURE], [TAG], etc.
    tokenize_method:
        Which text tokenizer to use.
        Must be consistent with TOKENIZER dictionary.

    Returns
    ----------

    Examples
    ----------
    >>> tokenizer = TokenizerForHuggingface(add_special_tokens=True)
    >>> item = "有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"
    >>> token_item = tokenizer(item)
    >>> print(token_item.input_ids[:10])
    tensor([[ 101, 1062, 2466, 1963, 1745,  138,  100,  140,  166,  117,  167, 5276,
             3338, 3340,  816, 1062, 2466,  102,  168,  134,  166,  116,  128,  167,
             3297, 1920,  966,  138,  100,  140,  102]])
    >>> print(tokenizer.tokenize(item)[:10])
    ['公', '式', '如', '图', '[', '[UNK]', ']', 'x', ',', 'y']
    >>> items = [item, item]
    >>> token_items = tokenizer(items, return_tensors='pt')
    >>> print(token_items.input_ids.shape)
    torch.Size([2, 31])
    >>> print(len(tokenizer.tokenize(items)))
    2
    >>> tokenizer.save_pretrained('test_dir') # doctest: +SKIP
    >>> tokenizer = TokenizerForHuggingface.from_pretrained('test_dir') # doctest: +SKIP
    """
    def __init__(self, pretrained_model="bert-base-chinese", max_length=512, tokenize_method: str = "pure_text",
                 add_specials: Union[List[str], bool] = False, **kwargs):
        self._set_basic_tokenizer(tokenize_method, **kwargs)
        if isinstance(add_specials, bool):
            add_specials = EDU_SPYMBOLS if add_specials is True else []
        else:
            add_specials = EDU_SPYMBOLS + add_specials
        self._special_tokens = set()
        self.max_length = max_length
        self.bert_tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.add_specials(add_specials)
        config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "pretrained_model", "kwargs"]}
        config.update(kwargs)
        self.config = config

    def __call__(self, items: Tuple[list, str, dict], key=lambda x: x, padding=True,
                 return_tensors: Optional[Tuple[str, TensorType, bool]] = True, **kwargs):
        if isinstance(items, list):
            text = [self._pre_tokenize(key(i)) for i in items]
        else:
            text = self._pre_tokenize(key(items))

        if isinstance(return_tensors, bool):
            return_tensors = "pt" if return_tensors is True else None
        encodes = self.bert_tokenizer(text, truncation=True, padding=padding, max_length=self.max_length,
                                      return_tensors=return_tensors)
        return encodes

    def __len__(self):
        return len(self.bert_tokenizer)

    def _set_basic_tokenizer(self, tokenize_method: str = None, **kwargs):
        self.tokenize_method = tokenize_method
        if self.tokenize_method is not None:
            self.text_tokenizer = get_tokenizer(tokenize_method, **kwargs)
        else:
            self.text_tokenizer = None

    def _pre_tokenize(self, text: Union[str, dict]):
        if self.text_tokenizer is not None:
            text = self.text_tokenizer._tokenize(text)
            text = " ".join(text)
        return text

    def tokenize(self, items: Union[list, str, dict], key=lambda x: x, **kwargs):
        if isinstance(items, list):
            texts = [self._tokenize(key(i)) for i in items]
            return texts
        else:
            return self._tokenize(key(items))

    def encode(self, items: Tuple[str, dict, List[str], List[dict]], key=lambda x: x, **kwargs):
        if isinstance(items, str) or isinstance(items, dict):
            return self.bert_tokenizer.encode(key(items), **kwargs)
        else:
            return [self.bert_tokenizer.encode(key(item), **kwargs) for item in items]

    def decode(self, token_ids: list, key=lambda x: x, **kwargs):
        if isinstance(token_ids[0], list):
            return [self.bert_tokenizer.decode(key(item), **kwargs) for item in token_ids]
        else:
            return self.bert_tokenizer.decode(key(token_ids), **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        item = self._pre_tokenize(key(item))
        return self.bert_tokenizer.tokenize(item, **kwargs)

    @classmethod
    def from_pretrained(cls, tokenizer_config_dir, **kwargs):
        custom_config_dir = os.path.join(tokenizer_config_dir, 'custom_config.json')
        if os.path.exists(custom_config_dir):
            with open(custom_config_dir, 'r') as f:
                custom_config = json.load(f)
                custom_config.update(kwargs)
            return cls(tokenizer_config_dir, **custom_config)
        else:
            return cls(tokenizer_config_dir, **kwargs)

    def save_pretrained(self, tokenizer_config_dir):
        self.bert_tokenizer.save_pretrained(tokenizer_config_dir)
        custom_config = self.config
        with open(os.path.join(tokenizer_config_dir, 'custom_config.json'), 'w') as f:
            json.dump(custom_config, f, indent=2)

    @property
    def vocab_size(self):
        return len(self.bert_tokenizer)

    def set_vocab(self, items: Tuple[List[str], List[dict]], key=lambda x: x, lower=False,
                  trim_min_count: int = 1, do_tokenize: bool = True):
        """
        Parameters
        -----------
        items: list
            can be the list of str, or list of dict
        key: function
            determine how to get the text of each item
        trim_min_count : int, optional
            the lower bound number for adding a word into vocabulary, by default 1
        do_tokenize : bool, optional
            wheather tokenize items before updating vocab, by default True
        """
        word2cnt = dict()
        for item in items:
            tokens = self._pre_tokenize(key(item)).split() if do_tokenize else key(item)
            for word in tokens:
                word = word.lower() if lower else word
                word2cnt[word] = word2cnt.get(word, 0) + 1
        remain_tokens = [w for w, c in word2cnt.items() if c >= trim_min_count]
        added_num = self.add_tokens(remain_tokens)
        return remain_tokens, added_num

    def add_specials(self, added_spectials: List[str]):
        for tok in added_spectials:
            self._special_tokens.add(tok)

        return self.bert_tokenizer.add_special_tokens({'additional_special_tokens': added_spectials})

    def add_tokens(self, added_tokens: List[str]):
        return self.bert_tokenizer.add_tokens(added_tokens)



# coding: utf-8
# 2021/5/18 @ tongshiwei
import logging
import jieba
from .stopwords import DEFAULT_STOPWORDS

jieba.setLogLevel(logging.INFO)


def is_chinese(word):
    """判断一个char或者string是否是汉字(串)"""
    for char in word:
        if char < u'\u4e00' or char > u'\u9fa5':
            return False
    return True


def tokenize(text, granularity="word", stopwords="default"):
    """
    Using jieba library to tokenize item by word or char.

    Parameters
    ----------
    text
    granularity
    stopwords: str, None or set

    Returns
    -------

    Examples
    --------
    >>> tokenize("三角函数是基本初等函数之一")
    ['三角函数', '初等', '函数']
    >>> tokenize("三角函数是基本初等函数之一", granularity="char")
    ['三', '角', '函', '数', '初', '等', '函', '数']
    """
    stopwords = DEFAULT_STOPWORDS if stopwords == "default" else stopwords
    stopwords = stopwords if stopwords is not None else {}
    if granularity == "word":
        return [token for token in jieba.cut(text) if token not in stopwords and token.strip()]
    elif granularity == "char":
        jieba_tokens = [token for token in jieba.cut(text) if token not in stopwords and token.strip()]
        # Use jieba_tokens to hangle sentence with mixed chinese and english.
        split_tokens = []
        for token in jieba_tokens:
            if is_chinese(token):
                split_tokens.extend(list(token))
            else:
                split_tokens.append(token)
        return split_tokens
    else:
        raise TypeError("Unknown granularity %s" % granularity)



# -*- coding: utf-8 -*-

import logging
import torch
from torch import nn
from torch.optim.adam import Adam
from torch.optim.lr_scheduler import StepLR
import os
import json
from gensim.models import KeyedVectors

from .modules import TextEncoder, AttnModel, ConceptEstimator, MIEstimator, DisenEstimator
from .utils import get_mask
from ..utils import set_device
from ..base_model import BaseModel
from transformers.modeling_outputs import ModelOutput
from transformers import PretrainedConfig


class DisenQNetOutput(ModelOutput):
    """
    Output type of [`DisenQNet`]

    Parameters
    ----------
    embed: Tensor of (batch_size, seq_len, hidden_size), word embedding
    k_hidden: Tensor of (batch_size, hidden_size) or None, concept representation of question
    i_hidden: Tensor of (batch_size, hidden_size) or None, individual representation of question
    """
    embeded: torch.FloatTensor = None
    k_hidden: torch.FloatTensor = None
    i_hidden: torch.FloatTensor = None


class DisenQNet(BaseModel):
    base_model_prefix = 'disenq'
    """
    DisenQNet question representation model

    Parameters
    ----------
    vocab_size: int
        size of vocabulary
    hidden_size: int
        size of word and question embedding
    dropout_rate: float
        dropout rate
    wv: torch.Tensor
        Tensor of (vocab_size, hidden_size) or None, initial word embedding, default = None
    """

    def __init__(self, vocab_size: int, hidden_size: int, dropout_rate: float, wv=None, **kwargs):
        super(DisenQNet, self).__init__()
        self.hidden_size = hidden_size
        self.encoder = TextEncoder(vocab_size, hidden_size, dropout_rate, wv=wv)
        self.k_model = AttnModel(hidden_size, dropout_rate)
        self.i_model = AttnModel(hidden_size, dropout_rate)
        self.dropout = nn.Dropout(p=dropout_rate)
        # config
        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "kwargs", 'wv']}
        self.config.update(kwargs)
        self.config['architecture'] = 'DisenQNet'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, seq_idx=None, seq_len=None, get_vk=True, get_vi=True) -> ModelOutput:
        """
        Parameters
        ----------
        seq_idx: Tensor of (batch_size, seq_len)
            word index
        seq_len: Tensor of (batch_size)
            valid sequence length of each batch
        get_vk: bool
            whether to return vk
        get_vi: bool
            whether to return vi

        Returns
        -------
        DisenQNetOutput
            - embed: Tensor of (batch_size, seq_len, hidden_size), word embedding
            - k_hidden: Tensor of (batch_size, hidden_size) or None, concept representation of question
            - i_hidden: Tensor of (batch_size, hidden_size) or None, individual representation of question
        """
        # embed: batch_size * seq_len * hidden_size
        # q_hidden: batch_size * hidden_size
        embed, q_hidden = self.encoder(seq_idx)
        # batch_size * seq_len, 0 means valid, 1 means pad
        mask = get_mask(seq_idx.size(1), seq_len)
        embed.masked_fill_(mask.unsqueeze(-1), 0)
        k_hidden, i_hidden = None, None
        q_hidden_dp = self.dropout(q_hidden)
        embed_dp = self.dropout(embed)
        # batch_size * hidden_size
        if get_vk:
            k_hidden, _ = self.k_model(q_hidden_dp, embed_dp, embed_dp, mask)
        if get_vi:
            i_hidden, _ = self.i_model(q_hidden_dp, embed_dp, embed_dp, mask)
        return DisenQNetOutput(
            embeded=embed,
            k_hidden=k_hidden,
            i_hidden=i_hidden
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config['vocab_size'],
                hidden_size=model_config['hidden_size'],
                dropout_rate=model_config['dropout_rate'],
            )


class DisenQNetForPreTrainingOutput(ModelOutput):
    """
    Output type of [`DisenQNetForPreTraining`]

    Parameters
    ----------
    loss
    embed: Tensor of (batch_size, seq_len, hidden_size), word embedding
    k_hidden: Tensor of (batch_size, hidden_size) or None, concept representation of question
    i_hidden: Tensor of (batch_size, hidden_size) or None, individual representation of question
    """
    loss: torch.FloatTensor = None
    embeded: torch.FloatTensor = None
    k_hidden: torch.FloatTensor = None
    i_hidden: torch.FloatTensor = None


class DisenQNetForPreTraining(BaseModel):
    base_model_prefix = 'disenq'

    def __init__(self, vocab_size, concept_size, hidden_size, dropout_rate, pos_weight,
                 w_cp, w_mi, w_dis, warmup, n_adversarial, wv=None, **kwargs):
        super(DisenQNetForPreTraining, self).__init__()
        self.disenq = DisenQNet(
            vocab_size=vocab_size,
            hidden_size=hidden_size,
            dropout_rate=dropout_rate,
            wv=wv,
            **kwargs)
        self.mi_estimator = MIEstimator(hidden_size, hidden_size * 2, dropout_rate)
        self.concept_estimator = ConceptEstimator(hidden_size, concept_size, pos_weight, dropout_rate)
        self.disen_estimator = DisenEstimator(hidden_size, dropout_rate)
        self.w_cp = w_cp
        self.w_mi = w_mi
        self.w_dis = w_dis
        self.hidden_size = hidden_size
        self.warming_up = False
        self.params = {
            "vocab_size": vocab_size,
            "concept_size": concept_size,
            "hidden_size": hidden_size,
            "dropout": dropout_rate,
            "pos_weight": pos_weight,
            "w_cp": w_cp,
            "w_mi": w_mi,
            "w_dis": w_dis,
            'warmup': warmup,
            'n_adversarial': n_adversarial,
        }
        self.modules = (self.disenq, self.mi_estimator, self.concept_estimator, self.disen_estimator)

        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "kwargs", 'wv']}
        self.config.update(kwargs)
        self.config['architecture'] = 'DisenQNetForPreTraining'
        self.config = PretrainedConfig.from_dict(self.config)

        model_params = list()
        for params in [list(self.disenq.parameters()),
                       list(self.mi_estimator.parameters()), list(self.concept_estimator.parameters())]:
            model_params.extend(params)
        self.model_params = model_params
        self.adv_params = list(self.disen_estimator.parameters())

    def forward(self, seq_idx=None, seq_len=None, concept=None) -> ModelOutput:
        # train enc
        outputs = self.disenq(seq_idx, seq_len)
        embed = outputs.embeded
        k_hidden = outputs.k_hidden
        i_hidden = outputs.i_hidden
        hidden = torch.cat((k_hidden, i_hidden), dim=-1)
        # max mi
        mi_loss = - self.mi_estimator(embed, hidden, seq_len)
        # min concept_loss
        cp_loss = self.concept_estimator(k_hidden, concept)
        if self.warming_up:
            loss = self.w_mi * mi_loss + self.w_cp * cp_loss
        else:
            # min dis
            dis_loss = self.disen_estimator(k_hidden, i_hidden)
            loss = self.w_mi * mi_loss + self.w_cp * cp_loss + self.w_dis * dis_loss
        return DisenQNetForPreTrainingOutput(
            loss=loss,
            embeded=embed,
            k_hidden=k_hidden,
            i_hidden=i_hidden
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config['vocab_size'],
                concept_size=model_config['concept_size'],
                hidden_size=model_config['hidden_size'],
                dropout_rate=model_config['dropout_rate'],
                pos_weight=model_config['pos_weight'],
                w_cp=model_config['w_cp'],
                w_mi=model_config['w_mi'],
                w_dis=model_config['w_dis'],
                warmup=model_config['warmup'],
                n_adversarial=model_config['n_adversarial'],
            )



import torch
from torch.nn.utils.rnn import pack_padded_sequence


def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)


class SeqBatch:
    def __init__(self, seqs, dtype=None, device=None):
        self.dtype = dtype
        self.device = device
        self.seqs = seqs
        self.lens = [len(x) for x in seqs]

        self.ind = argsort(self.lens)[::-1]
        self.inv = argsort(self.ind)
        self.lens.sort(reverse=True)
        self._prefix = [0]
        self._index = {}
        c = 0
        for i in range(self.lens[0]):
            for j in range(len(self.lens)):
                if self.lens[j] <= i:
                    break
                self._index[i, j] = c
                c += 1

    def packed(self):
        ind = torch.tensor(self.ind, dtype=torch.long, device=self.device)
        padded = self.padded()[0].index_select(1, ind)
        return pack_padded_sequence(padded, torch.tensor(self.lens))

    def padded(self, max_len=None, batch_first=False):
        seqs = [torch.tensor(s, dtype=self.dtype, device=self.device)
                if not isinstance(s, torch.Tensor) else s
                for s in self.seqs]
        if max_len is None:
            max_len = self.lens[0]
        seqs = [s[:max_len] for s in seqs]
        mask = [[1] * len(s) + [0] * (max_len - len(s)) for s in seqs]

        trailing_dims = seqs[0].size()[1:]
        if batch_first:
            out_dims = (len(seqs), max_len) + trailing_dims
        else:
            out_dims = (max_len, len(seqs)) + trailing_dims

        padded = seqs[0].new(*out_dims).fill_(0)
        for i, tensor in enumerate(seqs):
            length = tensor.size(0)
            # use index notation to prevent duplicate references to the tensor
            if batch_first:
                padded[i, :length, ...] = tensor
            else:
                padded[:length, i, ...] = tensor
        return padded, torch.tensor(mask).byte().to(self.device)

    def index(self, item):
        return self._index[item[0], self.inv[item[1]]]

    def invert(self, batch, dim=0):
        return batch.index_select(dim, torch.tensor(self.inv, device=self.device))



# coding: utf-8
# 2021/8/2 @ tongshiwei

import fire


from EduNLP.Vector.t2v import get_all_pretrained_models


def list_i2v():
    print("\n".join(get_all_pretrained_models()))


def cli():  # pragma: no cover
    fire.Fire({"i2v": list_i2v})



import torch
import numpy as np
import torch.nn as nn
from .modules import FeatureExtractor, ImageAE, MetaAE
from .util import SeqBatch
import json
import os
from pathlib import Path
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_packed_sequence, PackedSequence
from torchvision.transforms.functional import to_tensor
from transformers.modeling_outputs import ModelOutput
from transformers import PretrainedConfig
from ..base_model import BaseModel


class QuesNetOutput(ModelOutput):
    """
    Output type of [`DisenQNet`]

    Parameters
    ----------
    pack_embeded: Tensor of (batch_size, seq_len, hidden_size), word embedding
    hidden: Tensor of (batch_size, hidden_size) or None
    """
    pack_embeded: torch.FloatTensor = None
    embeded: torch.FloatTensor = None
    hidden: torch.FloatTensor = None


class QuesNet(BaseModel, FeatureExtractor):
    base_model_prefix = 'quesnet'

    def __init__(self, _stoi=None, meta='know_name', pretrained_embs: np.ndarray = None,
                 pretrained_image: nn.Module = None, pretrained_meta: nn.Module = None,
                 lambda_input=None,
                 feat_size=256, emb_size=256, rnn_type='LSTM', layers=4, **kwargs):
        BaseModel.__init__(self)
        FeatureExtractor.__init__(self, feat_size=feat_size)
        self.feat_size = feat_size
        self.rnn_size = feat_size // 2
        self.emb_size = emb_size
        self.vocab_size = len(_stoi['word'])

        self.stoi = _stoi
        self.itos = {v: k for k, v in self.stoi['word'].items()}
        # Encoder - Word
        self.we = nn.Embedding(self.vocab_size, emb_size)
        if pretrained_embs is not None:
            self.load_emb(pretrained_embs)
        # Encoder - Image
        self.ie = ImageAE(emb_size)
        if pretrained_image is not None:
            self.load_img(pretrained_image)
        # Encoder - Mata
        self.meta = meta
        self.meta_size = len(_stoi[self.meta])
        self.me = MetaAE(self.meta_size, emb_size)
        if pretrained_meta is not None:
            self.load_meta(pretrained_meta)
        if lambda_input is None:
            lambda_input = [1., 1., 1.]
        self.lambda_input = lambda_input
        self.proj_q = nn.Linear(feat_size, feat_size)
        self.proj_k = nn.Linear(feat_size, feat_size)
        self.proj_v = nn.Linear(feat_size, feat_size)
        self.dropout = nn.Dropout(0.2)
        self.rnn_type = rnn_type
        if rnn_type == 'GRU':
            self.rnn = nn.GRU(emb_size, self.rnn_size, layers,
                              bidirectional=True, batch_first=True)
            self.h0 = nn.Parameter(torch.rand(layers * 2, 1, self.rnn_size))
        elif rnn_type == 'LSTM':
            self.rnn = nn.LSTM(emb_size, self.rnn_size, layers,
                               bidirectional=True, batch_first=True)
            self.h0 = nn.Parameter(torch.rand(layers * 2, 1, self.rnn_size))
            self.c0 = nn.Parameter(torch.rand(layers * 2, 1, self.rnn_size))
        else:
            raise ValueError('quesnet only support GRU and LSTM now.')
        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "kwargs"]}
        # self.config.update(kwargs)
        self.config["architecture"] = 'quesnet'
        self.config = PretrainedConfig.from_dict(self.config)

    def init_h(self, batch_size):
        size = list(self.h0.size())
        size[1] = batch_size
        if self.config.rnn_type == 'GRU':
            return self.h0.expand(size).contiguous()
        else:
            return self.h0.expand(size).contiguous(), self.c0.expand(size).contiguous()

    def load_emb(self, emb):
        self.we.weight.detach().copy_(torch.from_numpy(emb))

    def load_img(self, img_layer: nn.Module):
        if self.config.emb_size != img_layer.emb_size:
            raise ValueError("Unmatched pre-trained ImageAE and embedding size")
        else:
            self.ie.load_state_dict(img_layer.state_dict())

    def load_meta(self, meta_layer: nn.Module):
        if self.config.emb_size != meta_layer.emb_size or self.meta_size != meta_layer.meta_size:
            raise ValueError("Unmatched pre-trained MetaAE and embedding size or meta size")
        else:
            self.me.load_state_dict(meta_layer.state_dict())

    def make_batch(self, data, device, pretrain=False):
        """Returns embeddings"""
        lembs = []
        rembs = []
        embs = []
        gt = []
        ans_input = []
        ans_output = []
        false_options = [[] for i in range(3)]
        for q in data:
            meta = torch.zeros(len(self.stoi[self.meta])).to(device)
            meta[q.labels.get(self.meta) or []] = 1
            _lembs = [self.we(torch.tensor([0], device=device)),
                      self.we(torch.tensor([0], device=device)),
                      self.me.enc(meta.unsqueeze(0)) * self.lambda_input[2]]
            _rembs = [self.me.enc(meta.unsqueeze(0)) * self.lambda_input[2]]
            _embs = [self.we(torch.tensor([0], device=device)),
                     self.we(torch.tensor([0], device=device)),
                     self.me.enc(meta.unsqueeze(0)) * self.lambda_input[2]]
            _gt = [torch.tensor([0], device=device), meta]
            for w in q.content:
                if isinstance(w, int):
                    word = torch.tensor([w], device=device)
                    item = self.we(word) * self.lambda_input[0]
                    _lembs.append(item)
                    _rembs.append(item)
                    _embs.append(item)
                    _gt.append(word)
                else:
                    im = to_tensor(w).to(device)
                    item = self.ie.enc(im.unsqueeze(0), detach_tensor=True) * self.lambda_input[1]
                    _lembs.append(item)
                    _rembs.append(item)
                    _embs.append(item)
                    _gt.append(im)
            _gt.append(torch.tensor([0], device=device))
            _rembs.append(self.we(torch.tensor([0], device=device)))
            _rembs.append(self.we(torch.tensor([0], device=device)))
            _embs.append(self.we(torch.tensor([0], device=device)))
            _embs.append(self.we(torch.tensor([0], device=device)))

            lembs.append(torch.cat(_lembs, dim=0))
            rembs.append(torch.cat(_rembs, dim=0))
            embs.append(torch.cat(_embs, dim=0))
            gt.append(_gt)

            ans_input.append([0] + q.answer)
            ans_output.append(q.answer + [0])

            for i, fo in enumerate(q.false_options):
                false_options[i].append([0] + fo)

        lembs = SeqBatch(lembs, device=device)
        rembs = SeqBatch(rembs, device=device)
        embs = SeqBatch(embs, device=device)
        ans_input = SeqBatch(ans_input, device=device)
        ans_output = SeqBatch(ans_output, device=device)
        false_opt_input = [SeqBatch(foi, device=device) for foi in false_options]

        length = sum(lembs.lens)
        words = []
        ims = []
        metas = []
        # p = lembs.packed().data
        wmask = torch.zeros(length, device=device).byte()
        imask = torch.zeros(length, device=device).byte()
        mmask = torch.zeros(length, device=device).byte()

        for i, _gt in enumerate(gt):
            for j, v in enumerate(_gt):
                ind = lembs.index((j, i))
                if v.size() == torch.Size([1]):  # word
                    words.append((v, ind))
                    wmask[ind] = 1
                elif v.dim() == 1:  # meta
                    metas.append((v.unsqueeze(0), ind))
                    mmask[ind] = 1
                else:  # img
                    ims.append((v.unsqueeze(0), ind))
                    imask[ind] = 1
        words = [x[0] for x in sorted(words, key=lambda x: x[1])]
        ims = [x[0] for x in sorted(ims, key=lambda x: x[1])]
        metas = [x[0] for x in sorted(metas, key=lambda x: x[1])]

        words = torch.cat(words, dim=0) if words else None
        ims = torch.cat(ims, dim=0) if ims else None
        metas = torch.cat(metas, dim=0) if metas else None
        if pretrain:
            return (
                lembs, rembs, words, ims, metas, wmask, imask, mmask,
                embs, ans_input, ans_output, false_opt_input
            )
        else:
            return embs

    def forward(self, inputs: SeqBatch):
        packed = inputs.packed()
        h = self.init_h(packed.batch_sizes[0])
        # y: (batch_size, seq_len, 2*rnn_size)
        y, _ = self.rnn(packed, h)

        hs, lens = pad_packed_sequence(y, batch_first=True)
        mask = [[1] * lens[i].item() + [0] * (lens[0] - lens[i]).item()
                for i in range(len(lens))]
        mask = torch.tensor(mask).byte().to(inputs.device)

        # hs: (B, S, D), mask: (B, S)
        q, k, v = self.proj_q(hs), self.proj_k(hs), self.proj_v(hs)

        scores = q @ k.transpose(-2, -1) / np.sqrt(k.size(-1))  # (B, S, S)
        if mask is not None:
            mask = mask.float()
            scores = scores - 1e9 * (1.0 - mask.unsqueeze(1))
        scores = self.dropout(F.softmax(scores, dim=-1))  # (B, S, S)
        h = (scores @ v).max(1)[0]  # (B, D)

        return QuesNetOutput(
            pack_embeded=y,
            embeded=hs,
            hidden=inputs.invert(h, 0)
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(_stoi=model_config["_stoi"],
                       feat_size=model_config["feat_size"],
                       emb_size=model_config["emb_size"],
                       rnn_type=model_config["rnn_type"],
                       layers=model_config["layers"])


class QuesNetForPreTrainingOutput(ModelOutput):
    """
    Output type of [`DisenQNet`]

    Parameters
    ----------
    hidden: Tensor of (batch_size, hidden_size) or None
    """
    loss: torch.FloatTensor = None
    embeded: torch.FloatTensor = None
    hidden: torch.FloatTensor = None


class QuesNetForPreTraining(BaseModel):
    base_model_prefix = 'quesnet'
    """Sequence-to-sequence feature extractor based on RNN. Supports different
    input forms and different RNN types (LSTM/GRU), """
    def __init__(self, _stoi=None, pretrained_embs: np.ndarray = None, pretrained_image: nn.Module = None,
                 pretrained_meta: nn.Module = None,
                 meta='know_name',
                 emb_size=256,
                 feat_size=512,
                 rnn_type='LSTM',
                 lambda_input=None,
                 lambda_loss=None,
                 layers=4, **kwargs):
        BaseModel.__init__(self)
        self.quesnet = QuesNet(_stoi, meta=meta, pretrained_embs=pretrained_embs, pretrained_image=pretrained_image,
                               pretrained_meta=pretrained_meta, lambda_input=lambda_input, feat_size=feat_size,
                               emb_size=emb_size, rnn_type=rnn_type, layers=layers, **kwargs)
        self.vocab_size = self.quesnet.vocab_size
        self.feat_size = self.quesnet.feat_size
        self.emb_size = self.quesnet.emb_size
        self.rnn_size = self.quesnet.rnn_size
        if lambda_loss is None:
            lambda_loss = [1., 1.]
        self.lambda_loss = lambda_loss
        # 预训练头 - HLM
        self.woutput = nn.Linear(self.feat_size, self.vocab_size)
        self.ioutput = nn.Linear(self.feat_size, self.emb_size)
        self.moutput = nn.Linear(self.feat_size, self.emb_size)

        self.lwoutput = nn.Linear(self.rnn_size, self.vocab_size)
        self.lioutput = nn.Linear(self.rnn_size, self.emb_size)
        self.lmoutput = nn.Linear(self.rnn_size, self.emb_size)

        self.rwoutput = nn.Linear(self.rnn_size, self.vocab_size)
        self.rioutput = nn.Linear(self.rnn_size, self.emb_size)
        self.rmoutput = nn.Linear(self.rnn_size, self.emb_size)
        # 预训练头 - QA
        self.ans_decode = nn.GRU(self.emb_size, self.feat_size, layers,
                                 batch_first=True)
        self.ans_output = nn.Linear(self.feat_size, self.vocab_size)
        self.ans_judge = nn.Linear(self.feat_size, 1)
        self.dropout = nn.Dropout(0.2)

        self.config = {k: v for k, v in locals().items() if k not in [
            "self", "__class__", "kwargs", "pretrained_embs", "pretrained_image", "pretrained_meta"]}
        # self.config.update(kwargs)
        self.config['architecture'] = 'quesnet'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, batch):
        left, right, words, ims, metas, wmask, imask, mmask, inputs, ans_input, ans_output, false_opt_input = batch

        # high-level loss
        outputs = self.quesnet(inputs)
        embeded = outputs.embeded
        h = outputs.hidden

        x = ans_input.packed()
        y, _ = self.ans_decode(PackedSequence(self.quesnet.we(x.data), x.batch_sizes),
                               h.repeat(self.config.layers, 1, 1))
        floss = F.cross_entropy(self.ans_output(y.data),
                                ans_output.packed().data)
        floss = floss + F.binary_cross_entropy_with_logits(self.ans_judge(y.data),
                                                           torch.ones_like(self.ans_judge(y.data)))
        for false_opt in false_opt_input:
            x = false_opt.packed()
            y, _ = self.ans_decode(PackedSequence(self.quesnet.we(x.data), x.batch_sizes),
                                   h.repeat(self.config.layers, 1, 1))
            floss = floss + F.binary_cross_entropy_with_logits(self.ans_judge(y.data),
                                                               torch.zeros_like(self.ans_judge(y.data)))
        loss = floss * self.lambda_loss[1]
        # low-level loss
        left_hid = self.quesnet(left).pack_embeded.data[:, :self.rnn_size]
        right_hid = self.quesnet(right).pack_embeded.data[:, self.rnn_size:]

        wloss = iloss = mloss = None

        if words is not None:
            lwfea = torch.masked_select(left_hid, wmask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            lout = self.lwoutput(lwfea)
            rwfea = torch.masked_select(right_hid, wmask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            rout = self.rwoutput(rwfea)
            out = self.woutput(torch.cat([lwfea, rwfea], dim=1))
            wloss = (F.cross_entropy(out, words) + F.cross_entropy(lout, words) + F.
                     cross_entropy(rout, words)) * self.quesnet.lambda_input[0] / 3
            wloss *= self.lambda_loss[0]
            loss = loss + wloss

        if ims is not None:
            lifea = torch.masked_select(left_hid, imask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            lout = self.lioutput(lifea)
            rifea = torch.masked_select(right_hid, imask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            rout = self.rioutput(rifea)
            out = self.ioutput(torch.cat([lifea, rifea], dim=1))
            iloss = (self.quesnet.ie.loss(ims, out) + self.quesnet.ie.loss(ims, lout) + self.quesnet.ie.
                     loss(ims, rout)) * self.quesnet.lambda_input[1] / 3
            iloss *= self.lambda_loss[0]
            loss = loss + iloss

        if metas is not None:
            lmfea = torch.masked_select(left_hid, mmask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            lout = self.lmoutput(lmfea)
            rmfea = torch.masked_select(right_hid, mmask.unsqueeze(1).bool()) \
                .view(-1, self.rnn_size)
            rout = self.rmoutput(rmfea)
            out = self.moutput(torch.cat([lmfea, rmfea], dim=1))
            mloss = (self.quesnet.me.loss(metas, out) + self.quesnet.me.loss(metas, lout) + self.quesnet.me.
                     loss(metas, rout)) * self.quesnet.lambda_input[2] / 3
            mloss *= self.lambda_loss[0]
            loss = loss + mloss

        return QuesNetForPreTrainingOutput(
            loss=loss,
            embeded=embeded,
            hidden=h
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                _stoi=model_config["_stoi"],
                emb_size=model_config["emb_size"],
                rnn=model_config["rnn"], lambda_input=model_config["lambda_input"],
                lambda_loss=model_config["lambda_loss"], layers=model_config["layers"],
                feat_size=model_config["feat_size"]
            )



from typing import Optional, Union, List, Dict, Any, Iterable, Tuple
import traceback
import torch
import os
import json
import pandas as pd
from datasets import Dataset as HFDataset, load_from_disk
from torch.utils.data import Dataset
from ..Tokenizer import get_tokenizer
from ..ModelZoo.utils import pad_sequence
from ..SIF import EDU_SPYMBOLS


__all__ = ["EduVocab", "EduDataset", "PretrainedEduTokenizer"]


class EduVocab(object):
    """The vocabulary container for a corpus.

    Parameters
    ----------
    vocab_path : str, optional
        vocabulary path to initialize this container, by default None
    corpus_items : List[str], optional
        corpus items to update this vocabulary, by default None
    bos_token : str, optional
        token representing for the start of a sentence, by default "[BOS]"
    eos_token : str, optional
        token representing for the end of a sentence, by default "[EOS]"
    pad_token : str, optional
        token representing for padding, by default "[PAD]"
    unk_token : str, optional
        token representing for unknown word, by default "[UNK]"
    specials : List[str], optional
        spacials tokens in vocabulary, by default None
    lower : bool, optional
        wheather to lower the corpus items, by default False
    trim_min_count : int, optional
        the lower bound number for adding a word into vocabulary, by default 1
    """
    def __init__(self, vocab_path: str = None, corpus_items: List[str] = None, bos_token: str = "[BOS]",
                 eos_token: str = "[EOS]", pad_token: str = "[PAD]", unk_token: str = "[UNK]",
                 specials: List[str] = None, lower: bool = False, trim_min_count: int = 1, **kwargs):
        super(EduVocab, self).__init__()

        self._tokens = []
        self.idx_to_token = dict()
        self.token_to_idx = dict()
        self.frequencies = dict()
        # 定义特殊词
        self.bos_token = bos_token
        self.eos_token = eos_token
        self.pad_token = pad_token
        self.unk_token = unk_token
        self._special_tokens = [self.pad_token, self.unk_token, self.bos_token, self.eos_token]

        if specials:
            self._special_tokens += specials
        for st in self._special_tokens:
            self._add(st)
        # 加载词典
        if vocab_path is not None:
            self.load_vocab(vocab_path)
        elif corpus_items is not None:
            self.set_vocab(corpus_items, lower, trim_min_count)

        self.bos_idx = self.token_to_idx[self.bos_token]
        self.eos_idx = self.token_to_idx[self.eos_token]
        self.pad_idx = self.token_to_idx[self.pad_token]
        self.unk_idx = self.token_to_idx[self.unk_token]

    def __len__(self):
        return len(self._tokens)

    @property
    def vocab_size(self):
        return len(self._tokens)

    @property
    def special_tokens(self):
        return self._special_tokens

    @property
    def tokens(self):
        return self._tokens

    def to_idx(self, token):
        """convert token to index"""
        return self.token_to_idx.get(token, self.unk_idx)

    def to_token(self, idx):
        """convert index to index"""
        return self.idx_to_token.get(idx, self.unk_token)

    def convert_sequence_to_idx(self, tokens, bos=False, eos=False):
        """convert sentence of tokens to sentence of indexs"""
        res = [self.to_idx(t) for t in tokens]
        if bos is True:
            res = [self.bos_idx] + res
        if eos is True:
            res = res + [self.eos_idx]
        return res

    def convert_sequence_to_token(self, idxs, **kwargs):
        """convert sentence of indexs to sentence of tokens"""
        return [self.to_token(i) for i in idxs]

    def set_vocab(self, corpus_items: List[str], lower: bool = False, trim_min_count: int = 1, silent=True):
        """Update the vocabulary with the tokens in corpus items

        Parameters
        ----------
        corpus_items : List[str], optional
            corpus items to update this vocabulary, by default None
        lower : bool, optional
            wheather to lower the corpus items, by default False
        trim_min_count : int, optional
            the lower bound number for adding a word into vocabulary, by default 1
        """
        word2cnt = dict()
        for item in corpus_items:
            for word in item:
                word = word.lower() if lower else word
                word2cnt[word] = word2cnt.get(word, 0) + 1
        words = [w for w, c in word2cnt.items() if c >= trim_min_count and w not in self._special_tokens]
        for token in words:
            self._add(token)
        if not silent:
            keep_word_cnts = sum(word2cnt[w] for w in words)
            all_word_cnts = sum(word2cnt.values())
            print(f"save words(trim_min_count={trim_min_count}): {len(words)}/{len(word2cnt)} = {len(words) / len(word2cnt):.4f}\
                  with frequency {keep_word_cnts}/{all_word_cnts}={keep_word_cnts / all_word_cnts:.4f}")

    def load_vocab(self, vocab_path: str):
        """Load the vocabulary from vocab_file

        Parameters
        ----------
        vocab_path : str
            path to save vocabulary file
        """
        with open(vocab_path, "r", encoding="utf-8") as file:
            self._tokens = file.read().strip().split('\n')
            self.token_to_idx = {token: idx for idx, token in enumerate(self._tokens)}
            self.idx_to_token = {idx: token for idx, token in enumerate(self._tokens)}

    def save_vocab(self, vocab_path: str):
        """Save the vocabulary into vocab_file

        Parameters
        ----------
        vocab_path : str
            path to save vocabulary file
        """
        with open(vocab_path, 'w', encoding='utf-8') as file:
            for i in range(self.vocab_size):
                token = self._tokens[i]
                file.write(f"{token}\n")

    def _add(self, token: str):
        if token not in self._tokens:
            idx = len(self._tokens)
            self._tokens.append(token)
            self.idx_to_token[idx] = token
            self.token_to_idx[token] = idx

    def add_specials(self, tokens: List[str]):
        """Add special tokens into vocabulary"""
        for token in tokens:
            if token not in self._special_tokens:
                self._special_tokens += [token]
                self._add(token)

    def add_tokens(self, tokens: List[str]):
        """Add tokens into vocabulary"""
        for token in tokens:
            self._add(token)


# to do: how to handle tokenizer with formulas or pictures.
class PretrainedEduTokenizer(object):
    """This base class is in charge of preparing the inputs for a model

    Parameters
    ----------
    vocab_path : str, optional
        _description_, by default None
    max_length : int, optional
        used to clip the sentence out of max_length, by default None
    tokenize_method : str, optional
        default: "space"
        - when text is already seperated by space, use "space"
        - when text is raw string format, use Tokenizer defined in get_tokenizer(), such as "pure_text" and "text"
    add_specials : Tuple[list, bool], optional
        by default None
        - For bool, it means whether to add EDU_SPYMBOLS to vocabulary
        - For list, it means the added special tokens besides EDU_SPYMBOLS
    """
    def __init__(self, vocab_path: str = None, max_length: int = 250, tokenize_method: str = "pure_text",
                 add_specials: Tuple[list, bool] = False, **kwargs):
        self._set_basic_tokenizer(tokenize_method, **kwargs)
        if isinstance(add_specials, bool):
            add_specials = EDU_SPYMBOLS if add_specials else []
        else:
            add_specials = EDU_SPYMBOLS + add_specials
        self.max_length = max_length
        self.vocab = EduVocab(vocab_path=vocab_path, specials=add_specials, **kwargs)

        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "vocab_path"]}

    def __call__(self, items: Tuple[list, str, dict], key=lambda x: x, padding: Tuple[bool, str] = True,
                 max_length=None, return_tensors=True, return_text=False, **kwargs) -> Dict[str, Any]:
        """
        Parameters
        ----------
        items: list or str or dict
            the question items
        key: function
            determine how to get the text of each item
        padding: bool
            whether to pad the seq_idx
        return_tensors: bool
            whether to return data as tensors (would ignore text tokens)
        return_text: bool
            whether to return text tokens

        Returns
        -------
        ret: dict
            {"seq_idx": None, "seq_len": None}
            or {"seq_token": None, seq_idx": None, "seq_len": None}.
            The shape of element is (batch, seq) or (batch,).

        Notes:
        -------
        Be Make sure Tokenizer output batched tensors by default
        """
        batch_max_length = None
        max_length = self.max_length if max_length is None else max_length
        if isinstance(padding, str):
            if padding == "max_length":
                batch_max_length = max_length
                padding = True
            elif padding == "longest":
                padding = True
            elif padding == "do_not_pad":
                padding = False
            else:
                raise ValueError("'padding' must be `bool` or `string` in ['max_length', 'longest', 'do_not_pad']")

        token_items = self.tokenize(items, key)
        if isinstance(items, dict) or isinstance(items, str):
            token_items = [token_items]
        if max_length is not None:
            token_items = [seq[:max_length] for seq in token_items]
        seqs = [self.vocab.convert_sequence_to_idx(token_item,
                                                   bos=kwargs.get("bos", False),
                                                   eos=kwargs.get("eos", False)) for token_item in token_items]
        lengths = [len(seq) for seq in seqs]
        ret = {
            "seq_idx": pad_sequence(seqs, pad_val=self.vocab.pad_idx, max_length=batch_max_length) if padding else seqs,
            "seq_len": lengths
        }
        if isinstance(items, dict) or isinstance(items, str):
            ret = {k: v[0] for k, v in ret.items()}
            token_items = token_items[0]
        if return_tensors:
            ret = {key: torch.as_tensor(val) for key, val in ret.items()}
        if return_text:
            ret["seq_token"] = token_items
        return ret

    def __len__(self):
        return len(self.vocab)

    def _set_basic_tokenizer(self, tokenize_method: str, **kwargs):
        self.tokenize_method = tokenize_method
        self.text_tokenizer = get_tokenizer(tokenize_method, **kwargs)

    def tokenize(self, items: Tuple[list, str, dict], key=lambda x: x, **kwargs):
        """
        Parameters
        ----------
        items: list or str or dict
            the question items
        key: function
            determine how to get the text of each item

        Returns
        -------
        tokens: list
            the token of items
        """
        if isinstance(items, str) or isinstance(items, dict):
            return self._tokenize(items, key=key)
        else:
            return [self._tokenize(item, key=key) for item in items]

    def encode(self, items: Tuple[str, dict, List[str], List[dict]], key=lambda x: x, **kwargs):
        if isinstance(items, str) or isinstance(items, dict):
            return self.vocab.convert_sequence_to_idx(self.tokenize(key(items)), **kwargs)
        else:
            return [self.vocab.convert_sequence_to_idx(self.tokenize(key(item)), **kwargs) for item in items]

    def decode(self, token_ids: list, key=lambda x: x, **kwargs):
        if isinstance(token_ids[0], list):
            return [self.vocab.convert_sequence_to_token(key(item), **kwargs) for item in token_ids]
        else:
            return self.vocab.convert_sequence_to_token(key(token_ids), **kwargs)

    def _pad(self):
        raise NotImplementedError

    def _tokenize(self, item: Tuple[str, dict], key=lambda x: x):
        token_item = self.text_tokenizer._tokenize(item, key=key)
        if len(token_item) == 0:
            token_item = [self.vocab.unk_token]
        if len(token_item) > self.max_length:
            token_item = token_item[:self.max_length]
        return token_item

    @classmethod
    def from_pretrained(cls, tokenizer_config_dir: str, **kwargs):
        """Load tokenizer from local files

        Parameters:
        -----------
        tokenizer_config_dir: str
            The dir path containing tokenizer_config.json and vocab.list
        """
        tokenizer_config_path = os.path.join(tokenizer_config_dir, "tokenizer_config.json")
        pretrained_vocab_path = os.path.join(tokenizer_config_dir, "vocab.txt")

        with open(tokenizer_config_path, "r", encoding="utf-8") as rf:
            tokenizer_config = json.load(rf)
            tokenizer_config.update(kwargs)
            return cls(
                vocab_path=pretrained_vocab_path,
                **tokenizer_config)

    def save_pretrained(self, tokenizer_config_dir: str):
        """Save tokenizer into local files

        Parameters:
        -----------
        tokenizer_config_dir: str
            save tokenizer params in `/tokenizer_config.json` and save words in `/vocab.list`
        """
        if not os.path.exists(tokenizer_config_dir):
            os.makedirs(tokenizer_config_dir, exist_ok=True)
        tokenizer_config_path = os.path.join(tokenizer_config_dir, "tokenizer_config.json")
        vocab_path = os.path.join(tokenizer_config_dir, "vocab.txt")
        self.vocab.save_vocab(vocab_path)

        with open(tokenizer_config_path, "w", encoding="utf-8") as wf:
            json.dump(self.config, wf, ensure_ascii=False, indent=2)

    @property
    def vocab_size(self):
        return len(self.vocab)

    def set_vocab(self, items: list, key=lambda x: x, lower: bool = False,
                  trim_min_count: int = 1, do_tokenize: bool = True):
        """Update the vocabulary with the tokens in corpus items

        Parameters
        ----------
        items: list
            can be the list of str, or list of dict
        key: function, optional
            determine how to get the text of each item
        lower : bool, optional
            wheather to lower the corpus items, by default False
        trim_min_count : int, optional
            the lower bound number for adding a word into vocabulary, by default 1
        do_tokenize : bool, optional
            wheather tokenize items before updating vocab, by default True

        Returns
        -------
        list
            token_items
        """
        token_items = self.tokenize(items, key) if do_tokenize else [key(item) for item in items]
        self.vocab.set_vocab(corpus_items=token_items, trim_min_count=trim_min_count, lower=lower)
        return token_items

    def add_specials(self, tokens):
        """Add special tokens into vocabulary"""
        self.vocab.add_specials(tokens)

    def add_tokens(self, tokens):
        """Add tokens into vocabulary"""
        self.vocab.add_tokens(tokens)


class EduDataset(Dataset):
    """The base class implements a Dataset, which package the `datasets.Dataset`
    and provide more convenience, including parallel preprocessing, offline loadding and so on.

    Parameters
    ----------
    tokenizer :
        PretrainedEduTokenizer or model-specific Pretrained Tokenizer
    ds_disk_path : HFDataset, optional
        the dataset_path to save dataset used by `datasets.Dataset`, by default None
    items : Union[List[dict], List[str]], optional
        input items to process, by default None
    stem_key : str, optional
        the content of items to process, by default "text"
    label_key : Optional[str], optional
        the labels of items to process, by default None
    feature_keys : Optional[List[str]], optional
        the additional features of items to remain, by default None
    num_processor : int, optional
        specific the number of cpus for parallel speedup, by default None
    """
    def __init__(self, tokenizer, ds_disk_path: HFDataset = None,
                 items: Union[List[dict], List[str]] = None,
                 stem_key: str = "text", label_key: Optional[str] = None,
                 feature_keys: Optional[List[str]] = None,
                 num_processor: int = None, **kwargs):
        self.tokenizer = tokenizer
        feature_keys = [] if feature_keys is None else feature_keys
        if items is not None:
            assert ds_disk_path is None
            if isinstance(items[0], dict):
                assert stem_key is not None
                raw_columns = set(items[0].keys())
            if isinstance(items[0], str):
                assert stem_key is None and label_key is None
                stem_key = "text"
                raw_columns = set([stem_key])
            work_columns = set([stem_key] + feature_keys + ([label_key] if label_key is not None else []))
            redundant_columns = raw_columns - work_columns
            # 在线预处理特征
            items = items if isinstance(items[0], dict) else [{"text": i} for i in items]
            df = pd.DataFrame(items)
            df.drop(columns=list(redundant_columns), inplace=True)
            self.ds = HFDataset.from_pandas(df)
            """Note: map will break down for super large data which is greater than 4GB """
            self.ds = self.ds.map(lambda sample: tokenizer(sample[stem_key], return_tensors=False),
                                  num_proc=num_processor,
                                  batched=True, batch_size=1000)
            remove_columns = [stem_key]
        else:
            # 离线加载工作特征
            assert ds_disk_path is not None
            self.ds = load_from_disk(ds_disk_path)
            reserve_columns = list(tokenizer("edunlp", return_tensors=False).keys())\
                + feature_keys + ([label_key] if label_key is not None else [])
            remove_columns = list(set(self.ds.column_names) - set(reserve_columns))

        # 工作特征
        self.work_ds = self.ds.remove_columns(remove_columns) if len(remove_columns) > 0 else self.ds
        if label_key is not None:
            self.work_ds = self.work_ds.rename_columns({
                label_key: "labels",
            })

    def __getitem__(self, index):
        return self.work_ds[index]

    def __len__(self):
        return self.work_ds.num_rows

    def to_disk(self, ds_disk_path):
        """Save the processed dataset into local files"""
        self.ds.save_to_disk(ds_disk_path)

    def collect_fn(self):
        raise NotImplementedError



import torch
import torch.nn as nn


class FeatureExtractor(nn.Module):
    def __init__(self, feat_size=512):
        super(FeatureExtractor, self).__init__()
        self.feat_size = feat_size

    def make_batch(self, data, device, pretrain=False):
        """Make batch from input data (python data / np arrays -> tensors)"""
        raise NotImplementedError

    def load_emb(self, emb):
        pass

    def forward(self, *input):
        raise NotImplementedError


class AE(nn.Module):
    factor = 1

    def enc(self, item, *args, **kwargs):
        return self.encoder(item, *args, **kwargs)

    def dec(self, item, *args, **kwargs):
        return self.decoder(item, *args, **kwargs)

    def loss(self, item, emb=None):
        if emb is None:
            emb = self(item)
            out = self.dec(emb)
        else:
            out = self.dec(emb)

        return self.recons_loss(out, item)

    def forward(self, item):
        return self.enc(item)


class ImageAE(AE):
    def __init__(self, emb_size):
        super().__init__()
        self.emb_size = emb_size
        self.recons_loss = nn.MSELoss()
        self._encoder = nn.Sequential(
            nn.Conv2d(1, 16, 3, stride=3),
            nn.ReLU(True),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(16, 32, 3, stride=2),
            nn.ReLU(True),
            nn.MaxPool2d(2, stride=1),
            nn.Conv2d(32, emb_size, 3, stride=2)
        )
        self._decoder = nn.Sequential(
            nn.ConvTranspose2d(emb_size // self.factor, 32, 3, stride=2),
            nn.ReLU(True),
            nn.ConvTranspose2d(32, 16, 5, stride=3, padding=1),
            nn.ReLU(True),
            nn.ConvTranspose2d(16, 8, 5, stride=3),
            nn.ReLU(True),
            nn.ConvTranspose2d(8, 1, 2, stride=2, padding=1),
            nn.Sigmoid()
        )

    def encoder(self, item, detach_tensor=False):
        return self._encoder(item).detach().view(item.size(0), -1) if detach_tensor else self._encoder(item).view(
            item.size(0), -1)

    def decoder(self, emb, detach_tensor=False):
        return self._decoder(emb[:, :, None, None]).detach() if detach_tensor else self._decoder(emb[:, :, None, None])


class MetaAE(AE):
    def __init__(self, meta_size, emb_size):
        super().__init__()
        self.emb_size = emb_size
        self.meta_size = meta_size
        self.recons_loss = nn.BCEWithLogitsLoss()
        self.encoder = nn.Sequential(nn.Linear(meta_size, emb_size),
                                     nn.ReLU(True))
        # error: inplace
        # nn.Linear(emb_size, emb_size)
        self.decoder = nn.Sequential(nn.Linear(emb_size // self.factor,
                                               emb_size),
                                     nn.ReLU(True),
                                     nn.Linear(emb_size, meta_size))



import torch
from EduNLP import logger
from typing import Union, List, Callable, Optional, Dict, Any, Tuple
from .components import PREPROCESSING_PIPES
from ..Pretrain import PretrainedEduTokenizer
from ..ModelZoo.base_model import BaseModel
from transformers.modeling_outputs import ModelOutput
from abc import ABC, abstractmethod

GenericTensor = Union["torch.Tensor", List["GenericTensor"]]


class PreProcessingPipeline(object):
    """
    A pipeline for tokenization processing.
    You should use it by calling pipeline('pre-process'), instead of itself directly.

    Parameters
    ----------
    pipe_names: `str` or `List[str]`, optional
        The quickly initialized pipeline components. For availabel pipes, check TOKENIZE_PIPES in `components`.
        To add componets more flexiblely with specific arguments or custom name, use `add_pipe`.

    Examples
    ----------
    >>> tkn = PreProcessingPipeline(['is_sif', 'to_sif', 'is_sif', 'seg_describe'])
    >>> tkn.add_pipe(name='seg', symbol='fm', before='seg_describe')
    >>> tkn.component_names
    ['is_sif', 'to_sif', 'is_sif', 'seg', 'seg_describe']
    >>> item = "如图所示，则三角形ABC的面积是_。"
    >>> tkn(item)
    False
    True
    {'t': 3, 'f': 1, 'g': 0, 'm': 1}
    ['如图所示，则三角形', '[FORMULA]', '的面积是', '[MARK]', '。']
    >>> tkn.rename_pipe(0, 'is_sif_lol')
    >>> tkn.add_pipe('to_sif', component=lambda x:x, first=True) # This won't succeed for the same name pipe exists
    >>> tkn.add_pipe('identify', component=lambda x:x, before=1)
    >>> tkn.component_names
    ['is_sif_lol', 'identify', 'to_sif', 'is_sif_lol', 'seg', 'seg_describe']
    """

    def __init__(self,
                 pipe_names: Optional[Union[List[str], str]] = None
                 ):
        self._preproc_components = {}
        self.component_pipeline = []
        if isinstance(pipe_names, list) and len(pipe_names) > 0:
            if any(comp_name not in PREPROCESSING_PIPES for comp_name in pipe_names):
                logger.error('Some components not existed!')
                raise ValueError
            for pipe_name in pipe_names:
                if pipe_name not in self._preproc_components:
                    self._preproc_components[pipe_name] = PREPROCESSING_PIPES[pipe_name]()
            self.component_pipeline += [comp_name for comp_name in pipe_names]

    def __call__(self, inputs):
        for name in self.component_pipeline:
            proc = self._preproc_components[name]
            try:
                inputs = proc(inputs)
            except Exception as e:
                logger.error(e)
        return inputs

    def __len__(self):
        return len(self.component_pipeline)

    def _get_pipe_index(
            self,
            before: Optional[Union[str, int]] = None,
            after: Optional[Union[str, int]] = None,
            first: Optional[bool] = None,
            last: Optional[bool] = None
    ):
        if sum(arg is not None for arg in [before, after, first, last]) > 1:
            logger.error('Only one of before/after/first/last can be set!')
            raise ValueError
        if last or not any(arg is not None for arg in [before, after, first]):
            return len(self)
        elif first:
            return 0
        elif isinstance(before, str):
            if before not in self.component_pipeline:
                logger.error('The before pipe does not exists!')
                raise ValueError
            else:
                return self.component_pipeline.index(before)
        elif isinstance(before, int):
            if before < 0 or before >= len(self.component_pipeline):
                logger.error('The before index must be greater than 0 and less than current length!')
                raise ValueError
            else:
                return before
        elif isinstance(after, str):
            if after not in self.component_pipeline:
                logger.error('The after pipe does not exists!')
                raise ValueError
            else:
                return self.component_pipeline.index(after) + 1
        elif isinstance(after, int):
            if after < 0 or after >= len(self.component_pipeline):
                logger.error('The after index must be greater than 0 and less than current length!')
                raise ValueError
            else:
                return after + 1
        else:
            raise ValueError

    def add_pipe(
            self,
            name: str,
            component: Optional[Callable] = None,
            before: Optional[Union[str, int]] = None,
            after: Optional[Union[str, int]] = None,
            first: Optional[bool] = None,
            last: Optional[bool] = None,
            *args,
            **kwargs
    ):
        """
        Add a component to the tokenization pipeline.
        Valid component must be Callable and feat its next component. Only one parameter of before/after/first/last
        can be set. Default setting is `last`.
        Notice:
        1. Please try to avoid more than one usages of one same pipe, otherwise you can only modify them with index.
            i.e. `before` and `after` works well only when the pipe is unique.
        2. The `*args, **kwargs` parameters will be passed to component constructor in `PREPROCESSING_PIPES`,
            and this only works when you do not give a callable component.

        Parameters
        ----------
        name: `str`, required
            the name of pipe
        component: `Callable`, optional
            the custom pipe component, be careful with its nearest components' input&output.
        before: `str` or `int`, optional
            name or index of the component to insert new component directly before. Index start from 0.
        after: `str` or `int`, optional
            name or index of the component to insert new component directly after. Index start from 0.
        first: `bool`, optional
            if true, insert the component first in the pipeline.
        last: `bool`, optional
            if true, insert the component last in the pipeline.
        """
        pipe_index = self._get_pipe_index(before, after, first, last)
        if component is None and name not in self._preproc_components:
            if name not in PREPROCESSING_PIPES:
                logger.error(f'Unknown pipe "{name}"')
                raise ValueError
            else:
                self._preproc_components[name] = PREPROCESSING_PIPES[name](*args, **kwargs)
        else:
            if name in self._preproc_components:
                logger.warn(f'One preserved component "{name}" has the same name, inserting is stopped.')
                return
            self._preproc_components[name] = component
        self.component_pipeline.insert(pipe_index, name)

    def remove_pipe(
            self,
            pipe: Union[str, int]
    ):
        """
        Remove a component from the pre-processing pipeline
        """
        if isinstance(pipe, str):
            if pipe not in self._preproc_components:
                logger.error(f'Unknown pipe "{pipe}"')
                raise ValueError
            self.component_pipeline.remove(pipe)
            return self._preproc_components[pipe]
        else:
            removed = self.component_pipeline.pop(pipe)
            return self._preproc_components[removed]

    def rename_pipe(
            self,
            old_pipe: Union[str, int],
            new_name: str,
    ):
        """
        Rename a component from the pre-processing pipeline.

        Parameters
        ----------
        old_pipe: `str` or `int`, required
            old component name for `str`, or old component index in the pipeline for `int`
        new_name: `str`, required
            new name for the component
        """
        if isinstance(old_pipe, int):
            old_pipe = self.component_pipeline[old_pipe]
        if old_pipe not in self._preproc_components:
            logger.error(f'Unknown pipe "{old_pipe}"')
            raise ValueError
        self._preproc_components[new_name] = self._preproc_components.pop(old_pipe)
        self.component_pipeline = [new_name if i == old_pipe else i for i in self.component_pipeline]

    @property
    def component_names(self):
        """
        Get the names of pipeline components
        """
        return self.component_pipeline.copy()

    @property
    def pipeline(self):
        """
        Get the processing pipeline consisting of (name, component) tuples.
        """
        return [(name, self._preproc_components[name]) for name in self.component_pipeline]


class Pipeline(ABC):
    """
    The pipeline class is the class from which all pipelines inherit. Pipeline workflow is defined as a sequence of the
    following operations:

        Input -> PreProcessingPipeline -> Tokenization -> Model Inference
                                                    -> Post-Processing (downstream task dependent) -> Output

    This class is not for using directly, refer to `Pipeline.__init__.pipeline` function.
    """

    def __init__(self,
                 task: Optional[str] = None,
                 model: Optional[BaseModel] = None,
                 tokenizer: Optional[PretrainedEduTokenizer] = None,
                 preproc_pipe_names: Optional[List] = None,
                 **kwargs
                 ):
        if preproc_pipe_names is None:
            preproc_pipe_names = []
        self.preproc_pipeline = PreProcessingPipeline(preproc_pipe_names)
        self.tokenizer = tokenizer
        self.model = model
        self.task = task

        self._tokenize_params, self._forward_params, self._postprocess_params = self._sanitize_parameters(**kwargs)

    def __len__(self):
        _length = len(self.preproc_pipeline) + sum(
            component is not None for component in [self.tokenizer, self.model, self.task])
        return _length

    @abstractmethod
    def _sanitize_parameters(self, **pipeline_parameters: Dict):
        """
        _sanitize_parameters will be automatically called with any excessive named arguments
        from either '__init__' and '__call__' methods.
        Any inheritor of Pipeline should implement it, and it should return 3 dictionaries of parameters used by
        '_tokenize', '_forward' and 'postprocess' methods
        """
        raise NotImplementedError("_sanitize_parameters not implemented")

    @abstractmethod
    def _tokenize(self, input_: Any, **tokenize_parameters: Dict) -> Dict[str, GenericTensor]:
        """
        _tokenize will take the `input_` of a specific pipeline, go through it on a tokenizer and return a dictionary
        of everything necessary for `forward` to run properly.
        """
        raise NotImplementedError("_tokenize not implemented")

    @abstractmethod
    def _forward(self, input_: Dict[str, GenericTensor], **forward_parameters: Dict) -> ModelOutput:
        """
        _forward will receive the prepared dictionary from `tokenization` and run it on the model.
        """
        raise NotImplementedError("_forward not implemented")

    @abstractmethod
    def postprocess(self, model_outputs: ModelOutput, **postprocess_parameters: Dict) -> Any:
        """
        postprocess will receive the outputs of `_forward` method and reformat them into something more friendly
        based on specific task.
        """
        raise NotImplementedError("postprocess not implemented")

    def add_pipe(self, *args, **kwargs):
        """
        refer to PreProcessingPipeline.add_pipe
        """
        return self.preproc_pipeline.add_pipe(*args, **kwargs)

    def remove_pipe(self, *args, **kwargs):
        """
        refer to PreProcessingPipeline.remove_pipe
        """
        return self.preproc_pipeline.remove_pipe(*args, **kwargs)

    def rename_pipe(self, *args, **kwargs):
        """
        refer to PreProcessingPipeline.rename_pipe
        """
        return self.preproc_pipeline.rename_pipe(*args, **kwargs)

    @property
    def component_names(self):
        """
        Get the names of pipeline components
        """
        _component_names = self.preproc_pipeline.component_names.copy() if len(self.preproc_pipeline) > 0 else []
        if self.tokenizer is not None:
            _component_names.append("tokenizer")
        if self.model is not None:
            _component_names.append(self.model.__class__.__name__)
        if self.task is not None:
            _component_names.append(self.task)
        return _component_names

    @property
    def pipeline(self):
        """
        Get the processing pipeline consisting of (name, component) tuples.
        """
        _pipeline = self.preproc_pipeline.pipeline if len(self.preproc_pipeline) > 0 else []
        if self.tokenizer is not None:
            _pipeline.append(("tokenizer", self.tokenizer))
        if self.model is not None:
            _pipeline.append((self.model.__class__.__name__, self.model))
        if self.task is not None:
            _pipeline.append((self.task, None))
        return _pipeline

    def __call__(self, inputs, *args, num_workers=None, **kwargs):
        if args:
            logger.warning(f"Ignoring args: {args}")
        is_batch = isinstance(inputs, list)

        tokenize_params, forward_params, postprocess_params = self._sanitize_parameters(**kwargs)
        tokenize_params = {**self._tokenize_params, **tokenize_params}
        forward_params = {**self._forward_params, **forward_params}
        postprocess_params = {**self._postprocess_params, **postprocess_params}

        if is_batch:
            return [self.run_single(item, tokenize_params, forward_params, postprocess_params) for item in inputs]
        else:
            return self.run_single(inputs, tokenize_params, forward_params, postprocess_params)

    def run_single(self, inputs, tokenize_params, model_params, postprocess_params):
        tokenize_inputs = self.preproc_pipeline(inputs)
        model_inputs = self._tokenize([tokenize_inputs], **tokenize_params)
        model_outputs = self._forward(model_inputs, **model_params)
        outputs = self.postprocess(model_outputs, **postprocess_params)
        return outputs



import torch

from .base import Pipeline, GenericTensor
from typing import Dict, Optional, Union
from torch import sigmoid


class KnowledgePredictionPipeline(Pipeline):
    def __init__(self, **kwargs):
        super(KnowledgePredictionPipeline, self).__init__(**kwargs)

    def _sanitize_parameters(self, **pipeline_parameters):
        tokenize_params, forward_params, postprocess_params = pipeline_parameters, {}, {}
        return tokenize_params, forward_params, postprocess_params

    def _tokenize(self, input_, **tokenize_parameters) -> Dict[str, GenericTensor]:
        return self.tokenizer(input_, **tokenize_parameters)

    def _forward(self, model_inputs, **forward_params):
        return self.model(**model_inputs)

    def postprocess(self, model_outputs, **postprocess_params):
        if 'num_classes_list' not in dir(self.model) or 'num_total_classes' not in dir(self.model):
            raise ValueError('model is not for knowledge prediction: ', self.model)
        outputs = model_outputs["logits"][0]
        start_idx = 0
        knowledge_list = []
        for num_classes in self.model.num_classes_list:
            level_prediction = torch.argmax(outputs[start_idx:start_idx + num_classes]) + start_idx
            knowledge_list.append(level_prediction)
            start_idx += num_classes
        outputs = outputs.detach().numpy()
        dict_knowledge = {
            "knowledge_list": knowledge_list,
            "knowledge_scores": outputs.tolist(),
        }
        return dict_knowledge



from .base import Pipeline, GenericTensor
from typing import Dict, Optional, Union


class PropertyPredictionPipeline(Pipeline):
    def __init__(self, **kwargs):
        super(PropertyPredictionPipeline, self).__init__(**kwargs)

    def _sanitize_parameters(self, **pipeline_parameters):
        tokenize_params, forward_params, postprocess_params = pipeline_parameters, {}, {}
        return tokenize_params, forward_params, postprocess_params

    def _tokenize(self, input_, **tokenize_parameters) -> Dict[str, GenericTensor]:
        return self.tokenizer(input_, **tokenize_parameters)

    def _forward(self, model_inputs, **forward_params):
        return self.model(**model_inputs)

    def postprocess(self, model_outputs, **postprocess_params):
        outputs = model_outputs["logits"]
        outputs = outputs.detach().numpy()
        dict_property = {"property": outputs.item()}
        return dict_property



from pathlib import PurePath
import os
import torch
import torch.optim as optim
import torch.nn as nn
import torch.utils.data as tud
import numpy as np
from torch.nn.utils.rnn import pad_sequence
import torch.nn.functional as F
from EduNLP.Pretrain import train_elmo, ElmoTokenizer
from EduNLP.ModelZoo.rnn import ElmoLM
from .meta import Vector
import json
from typing import Dict, List, Tuple


class ElmoModel(Vector):
    def __init__(self, pretrained_dir: str):
        """
        Parameters
        ----------
        pretrained_model_path: str
        """
        super(ElmoModel, self).__init__()
        self.model = ElmoLM.from_pretrained(pretrained_dir)
        self.model.eval()

    def __call__(self, *args, **kwargs):
        return self.infer_vector(*args, **kwargs)

    def infer_vector(self, items: Tuple[dict, List[dict]], *args, **kwargs) -> torch.Tensor:
        # TODO: handle batch and unbatch format for inputs and outputs
        # is_batch = isinstance(items, list)
        # items = items if is_batch else [items]
        outputs = self.model(**items)
        item_embeds = torch.cat(
            (outputs.forward_output[torch.arange(len(items["seq_len"])), torch.tensor(items["seq_len"]) - 1],
             outputs.backward_output[torch.arange(len(items["seq_len"])), 0]),
            dim=-1)
        return item_embeds

    def infer_tokens(self, items, *args, **kwargs) -> torch.Tensor:
        # is_batch = isinstance(items, list)
        outputs = self.model(**items)
        forward_hiddens = outputs.forward_output
        backward_hiddens = outputs.backward_output
        return torch.cat((forward_hiddens, backward_hiddens), dim=-1)
        # if is_batch:
        #     ret = []
        #     for fh, bh, lg in zip(forward_hiddens, backward_hiddens, items.seq_len):
        #         _bh = torch.cat((torch.flip(bh[:lg], [0]), bh[lg:]), dim=0)
        #         ret.append(torch.cat((fh, _bh), dim=-1))
        #     return torch.stack(tuple(ret))
        # else:
        #     return torch.cat((forward_hiddens[0], torch.flip(backward_hiddens, [1])[0]), dim=-1)

    @property
    def vector_size(self):
        return 2 * self.model.hidden_size



import torch
from torch import nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
from baize.torch import load_net
import torch.nn.functional as F
import json
import os
from typing import List
from transformers.modeling_outputs import ModelOutput
from transformers import PretrainedConfig
from typing import Optional
from ..base_model import BaseModel
from ..utils import torch_utils as mytorch
from .harnn import HAM

__all__ = ["LM", "ElmoLM", "ElmoLMForPreTraining", "ElmoLMForPropertyPrediction", "ElmoLMForKnowledgePrediction"]


class LM(nn.Module):
    """

    Parameters
    ----------
    rnn_type：str
        Legal types including RNN, LSTM, GRU, BiLSTM
    vocab_size: int
    embedding_dim: int
    hidden_size: int
    num_layers
    bidirectional
    embedding
    model_params
    kwargs

    Examples
    --------
    >>> import torch
    >>> seq_idx = torch.LongTensor([[1, 2, 3], [1, 2, 0], [3, 0, 0]])
    >>> seq_len = torch.LongTensor([3, 2, 1])
    >>> lm = LM("RNN", 4, 3, 2)
    >>> output, hn = lm(seq_idx, seq_len)
    >>> output.shape
    torch.Size([3, 3, 2])
    >>> hn.shape
    torch.Size([1, 3, 2])
    >>> lm = LM("RNN", 4, 3, 2, num_layers=2)
    >>> output, hn = lm(seq_idx, seq_len)
    >>> output.shape
    torch.Size([3, 3, 2])
    >>> hn.shape
    torch.Size([2, 3, 2])
    """

    def __init__(self, rnn_type: str, vocab_size: int, embedding_dim: int, hidden_size: int, num_layers=1,
                 bidirectional=False, embedding=None, model_params=None, use_pack_pad=True, **kwargs):
        super(LM, self).__init__()
        rnn_type = rnn_type.upper()
        self.embedding = torch.nn.Embedding(vocab_size, embedding_dim) if embedding is None else embedding
        self.use_pack_pad = use_pack_pad
        self.c = False
        if rnn_type == "RNN":
            self.rnn = torch.nn.RNN(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
        elif rnn_type == "LSTM":
            self.rnn = torch.nn.LSTM(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
            self.c = True
        elif rnn_type == "GRU":
            self.rnn = torch.nn.GRU(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
        elif rnn_type == "BILSTM":
            bidirectional = True
            self.rnn = torch.nn.LSTM(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
            self.c = True
        else:
            raise TypeError("Unknown rnn_type %s" % rnn_type)

        self.num_layers = num_layers
        self.bidirectional = bidirectional
        if bidirectional is True:
            self.num_layers *= 2
        self.hidden_size = hidden_size
        if model_params:
            load_net(model_params, self, allow_missing=True)

    def forward(self, seq_idx, seq_len):
        """

        Parameters
        ----------
        seq_idx:Tensor
            a list of indices
        seq_len:Tensor
            length
        Returns
        --------
        sequence
            a PackedSequence object
        """
        seq = self.embedding(seq_idx)
        if self.use_pack_pad:
            seq_or_pack = pack_padded_sequence(seq, seq_len.cpu(), batch_first=True, enforce_sorted=False)
        else:
            seq_or_pack = seq
        h0 = torch.zeros(self.num_layers, seq.shape[0], self.hidden_size).to(seq_idx.device)
        if self.c is True:
            c0 = torch.zeros(self.num_layers, seq.shape[0], self.hidden_size).to(seq_idx.device)
            output, (hn, _) = self.rnn(seq_or_pack, (h0, c0))
        else:
            output, hn = self.rnn(seq_or_pack, h0)
        if self.use_pack_pad:
            output, _ = pad_packed_sequence(output, batch_first=True)

        return output, hn


class ElmoLMOutput(ModelOutput):
    """
    Output type of [`ElmoLM`]

    Parameters
    ----------
    pred_forward: of shape (batch_size, sequence_length)
    pred_backward: of shape (batch_size, sequence_length)
    forward_output: of shape (batch_size, sequence_length, hidden_size)
    backward_output: of shape (batch_size, sequence_length, hidden_size)
    """
    pred_forward: torch.FloatTensor = None
    pred_backward: torch.FloatTensor = None
    forward_output: torch.FloatTensor = None
    backward_output: torch.FloatTensor = None


class ElmoLM(BaseModel):
    base_model_prefix = 'elmo'

    def __init__(self, vocab_size: int, embedding_dim: int, hidden_size: int, num_layers: int = 2,
                 dropout_rate: float = 0.5, use_pack_pad=False, **kwargs):
        super(ElmoLM, self).__init__()
        self.LM_layer = LM("BiLSTM", vocab_size, embedding_dim, hidden_size, num_layers=num_layers,
                           use_pack_pad=use_pack_pad, **kwargs)
        self.pred_layer = nn.Linear(hidden_size, vocab_size)
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.hidden_size = hidden_size
        self.dropout = nn.Dropout(dropout_rate)
        self.config = {k: v for k, v in locals().items() if k != "self" and k != "__class__" and k != "kwargs"}
        self.config.update(kwargs)
        self.config['architecture'] = 'ElmoLM'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, seq_idx=None, seq_len=None) -> ModelOutput:
        """
        Parameters
        ----------
        seq_idx:Tensor, of shape (batch_size, sequence_length)
            a list of indices
        seq_len:Tensor, of shape (batch_size)
            length

        Returns
        ----------
        ElmoLMOutput
            pred_forward: of shape (batch_size, sequence_length)
            pred_backward: of shape (batch_size, sequence_length)
            forward_output: of shape (batch_size, sequence_length, hidden_size)
            backward_output: of shape (batch_size, sequence_length, hidden_size)
        """
        lm_output, _ = self.LM_layer(seq_idx, seq_len)
        forward_output = lm_output[:, :, :self.hidden_size]
        backward_output = lm_output[:, :, self.hidden_size:]
        forward_output = self.dropout(forward_output)
        backward_output = self.dropout(backward_output)
        pred_forward = self.pred_layer(forward_output)
        pred_backward = self.pred_layer(backward_output)

        return ElmoLMOutput(
            pred_forward=pred_forward,
            pred_backward=pred_backward,
            forward_output=forward_output,
            backward_output=backward_output
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config['vocab_size'],
                embedding_dim=model_config['embedding_dim'],
                hidden_size=model_config['hidden_size'],
                dropout_rate=model_config['dropout_rate'],
                batch_first=model_config['batch_first']
            )


class ElmoLMForPreTrainingOutput(ModelOutput):
    """
    Output type of [`ElmoLMForPreTraining`].

    Parameters
    ----------
    loss:
    pred_forward: of shape (batch_size, sequence_length)
    pred_backward: of shape (batch_size, sequence_length)
    forward_output: of shape (batch_size, sequence_length, hidden_size)
    backward_output: of shape (batch_size, sequence_length, hidden_size)
    """
    loss: torch.FloatTensor = None
    pred_forward: torch.FloatTensor = None
    pred_backward: torch.FloatTensor = None
    forward_output: torch.FloatTensor = None
    backward_output: torch.FloatTensor = None


class ElmoLMForPreTraining(BaseModel):
    base_model_prefix = 'elmo'

    def __init__(self, vocab_size: int, embedding_dim: int, hidden_size: int, dropout_rate: float = 0.5,
                 batch_first=True, use_pack_pad=False, **kwargs):
        super(ElmoLMForPreTraining, self).__init__()
        self.elmo = ElmoLM(
            vocab_size=vocab_size,
            embedding_dim=embedding_dim,
            hidden_size=hidden_size,
            dropout_rate=dropout_rate,
            batch_first=batch_first,
            use_pack_pad=use_pack_pad,
            **kwargs
        )
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.hidden_size = hidden_size
        self.criterion = nn.CrossEntropyLoss()

        self.config = {k: v for k, v in locals().items() if k != "self" and k != "__class__" and k != "kwargs"}
        self.config.update(kwargs)
        self.config['architecture'] = 'ElmoLMForPreTraining'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, seq_idx=None, seq_len=None) -> ModelOutput:
        """

        Parameters
        ----------
        seq_idx:Tensor, of shape (batch_size, sequence_length)
            a list of indices
        seq_len:Tensor, of shape (batch_size)
            length
        pred_mask : Tensor, of shape(batch_size, sequence_length)
        idx_mask : Tensor, of shape (batch_size, sequence_length)

        Returns
        -------
        ElmoLMForPreTrainingOutput
            loss
            pred_forward: of shape (batch_size, sequence_length)
            pred_backward: of shape (batch_size, sequence_length)
            forward_output: of shape (batch_size, sequence_length, hidden_size)
            backward_output: of shape (batch_size, sequence_length, hidden_size)
        """
        batch_size, idx_len = seq_idx.shape
        max_len = seq_len.max().item() if self.config.use_pack_pad is True else idx_len
        # Note:
        # pred_mask matters when LM use pack_pad,
        # but it will break down for parallel GPU because of different seq_len between gpus
        pred_mask = torch.arange(max_len, device=seq_idx.device)[None, :] < seq_len[:, None]
        idx_mask = torch.arange(idx_len, device=seq_idx.device)[None, :] < seq_len[:, None]

        pred_forward_mask = pred_mask.clone()
        pred_forward_mask[torch.arange(batch_size).unsqueeze(1), seq_len.unsqueeze(1) - 1] = False
        pred_backward_mask = pred_mask.clone()
        pred_backward_mask[torch.arange(batch_size).unsqueeze(1), 0] = False

        idx_forward_mask = idx_mask.clone()
        idx_forward_mask[torch.arange(batch_size).unsqueeze(1), 0] = False
        idx_backward_mask = idx_mask.clone()
        idx_backward_mask[torch.arange(batch_size).unsqueeze(1), seq_len.unsqueeze(1) - 1] = False

        outputs = self.elmo(seq_idx, seq_len)
        pred_forward, pred_backward = outputs.pred_forward, outputs.pred_backward

        flat_pred_forward = pred_forward[pred_forward_mask]
        flat_pred_backward = pred_backward[pred_backward_mask]

        # _, flat_pred_idx_forward = torch.max(flat_pred_forward, dim=1)
        # _, flat_pred_idx_backward = torch.max(flat_pred_backward, dim=1)

        flat_y_backward = seq_idx[idx_backward_mask]
        flat_y_forward = seq_idx[idx_forward_mask]

        # diff_forward = torch.sum(flat_pred_idx_forward - flat_y_forward)
        # diff_backward = torch.sum(flat_pred_idx_backward - flat_y_backward)

        forward_loss = self.criterion(flat_pred_forward, flat_y_forward)
        backward_loss = self.criterion(flat_pred_backward, flat_y_backward)
        loss = forward_loss + backward_loss
        return ElmoLMForPreTrainingOutput(
            loss=loss,
            pred_forward=pred_forward,
            pred_backward=pred_backward,
            forward_output=outputs.forward_output,
            backward_output=outputs.backward_output
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config['vocab_size'],
                embedding_dim=model_config['embedding_dim'],
                hidden_size=model_config['hidden_size'],
                dropout_rate=model_config['dropout_rate'],
                batch_first=model_config['batch_first']
            )


class PropertyPredictionOutput(ModelOutput):
    loss: torch.FloatTensor = None
    logits: torch.FloatTensor = None


class ElmoLMForPropertyPrediction(BaseModel):
    base_model_prefix = 'elmo'

    def __init__(self, vocab_size: int, embedding_dim: int, hidden_size: int, dropout_rate: float = 0.5,
                 batch_first=True, head_dropout=0.5, **kwargs):
        super(ElmoLMForPropertyPrediction, self).__init__()

        self.elmo = ElmoLM(
            vocab_size=vocab_size,
            embedding_dim=embedding_dim,
            hidden_size=hidden_size,
            dropout_rate=dropout_rate,
            batch_first=batch_first
        )
        self.head_dropout = head_dropout
        self.dropout = nn.Dropout(head_dropout)
        self.classifier = nn.Linear(2 * hidden_size, 1)
        self.sigmoid = nn.Sigmoid()
        self.criterion = nn.MSELoss()

        self.config = {k: v for k, v in locals().items() if k != "self" and k != "__class__" and k != "kwargs"}
        self.config.update(kwargs)
        self.config['architecture'] = 'ElmoLMForPreTraining'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, seq_idx=None, seq_len=None, labels=None) -> ModelOutput:
        outputs = self.elmo(seq_idx, seq_len)
        item_embeds = torch.cat(
            (outputs.forward_output[torch.arange(len(seq_len)), torch.tensor(seq_len) - 1],
             outputs.backward_output[torch.arange(len(seq_len)), 0]),
            dim=-1)
        item_embeds = self.dropout(item_embeds)

        logits = self.sigmoid(self.classifier(item_embeds))
        loss = None
        if labels is not None:
            loss = self.criterion(logits, labels)
        return PropertyPredictionOutput(
            loss=loss,
            logits=logits
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config.get('vocab_size'),
                embedding_dim=model_config.get('embedding_dim'),
                hidden_size=model_config.get('hidden_size'),
                dropout_rate=model_config.get('dropout_rate'),
                batch_first=model_config.get('batch_first'),
                head_dropout=model_config.get('head_dropout', 0.5),
            )


class KnowledgePredictionOutput(ModelOutput):
    loss: torch.FloatTensor = None
    logits: torch.FloatTensor = None


class ElmoLMForKnowledgePrediction(BaseModel):
    base_model_prefix = 'elmo'

    def __init__(self, vocab_size: int,
                 embedding_dim: int,
                 hidden_size: int,
                 num_classes_list: List[int],
                 num_total_classes: int,
                 dropout_rate: float = 0.5,
                 batch_first=True,
                 head_dropout: Optional[float] = 0.5,
                 flat_cls_weight: Optional[float] = 0.5,
                 attention_unit_size: Optional[int] = 256,
                 fc_hidden_size: Optional[int] = 512,
                 beta: Optional[float] = 0.5,
                 **kwargs):
        super(ElmoLMForKnowledgePrediction, self).__init__()

        self.elmo = ElmoLM(
            vocab_size=vocab_size,
            embedding_dim=embedding_dim,
            hidden_size=hidden_size,
            dropout_rate=dropout_rate,
            batch_first=batch_first
        )
        self.head_dropout = head_dropout
        self.dropout = nn.Dropout(head_dropout)
        self.sigmoid = nn.Sigmoid()
        self.criterion = nn.MSELoss()
        self.flat_classifier = nn.Linear(in_features=2 * hidden_size, out_features=num_total_classes)
        self.ham_classifier = HAM(
            num_classes_list=num_classes_list,
            num_total_classes=num_total_classes,
            sequence_model_hidden_size=hidden_size * 2,
            attention_unit_size=attention_unit_size,
            fc_hidden_size=fc_hidden_size,
            beta=beta,
            dropout_rate=dropout_rate
        )
        self.flat_cls_weight = flat_cls_weight
        self.num_classes_list = num_classes_list
        self.num_total_classes = num_total_classes

        self.config = {k: v for k, v in locals().items() if k != "self" and k != "__class__" and k != "kwargs"}
        self.config.update(kwargs)
        self.config['architecture'] = 'ElmoLMForPreTraining'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self, seq_idx=None, seq_len=None, labels=None) -> ModelOutput:
        outputs = self.elmo(seq_idx, seq_len)
        item_embeds = torch.cat(
            (outputs.forward_output[torch.arange(len(seq_len)), torch.tensor(seq_len) - 1],
             outputs.backward_output[torch.arange(len(seq_len)), 0]),
            dim=-1)
        tokens_embeds = torch.cat((outputs.forward_output, outputs.backward_output), dim=-1)
        item_embeds = self.dropout(item_embeds)
        tokens_embeds = self.dropout(tokens_embeds)
        flat_logits = self.sigmoid(self.flat_classifier(item_embeds))
        ham_outputs = self.ham_classifier(tokens_embeds)
        ham_logits = self.sigmoid(ham_outputs.scores)
        logits = self.flat_cls_weight * flat_logits + (1 - self.flat_cls_weight) * ham_logits
        loss = None
        if labels is not None:
            labels = torch.sum(torch.nn.functional.one_hot(labels, num_classes=self.num_total_classes), dim=1)
            labels = labels.float()
            loss = self.criterion(logits, labels)
        return KnowledgePredictionOutput(
            loss=loss,
            logits=logits
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                vocab_size=model_config.get('vocab_size'),
                embedding_dim=model_config.get('embedding_dim'),
                hidden_size=model_config.get('hidden_size'),
                num_total_classes=model_config.get('num_total_classes'),
                num_classes_list=model_config.get('num_classes_list'),
                dropout_rate=model_config.get('dropout_rate'),
                batch_first=model_config.get('batch_first'),
                head_dropout=model_config.get('head_dropout', 0.5),
                flat_cls_weight=model_config.get('flat_cls_weight', 0.5),
                attention_unit_size=model_config.get('attention_unit_size', 256),
                fc_hidden_size=model_config.get('fc_hidden_size', 512),
                beta=model_config.get('beta', 0.5),
            )



import os
from typing import List, Union
from transformers import BertForMaskedLM
from transformers import DataCollatorForLanguageModeling, DataCollatorWithPadding
from transformers import Trainer, TrainingArguments
from copy import deepcopy

from ..ModelZoo.bert import BertForPropertyPrediction, BertForKnowledgePrediction
from .pretrian_utils import EduDataset
from .hugginface_utils import TokenizerForHuggingface

__all__ = ["BertTokenizer", "BertDataset", "finetune_bert", "finetune_bert_for_property_prediction",
           "finetune_bert_for_knowledge_prediction"]

DEFAULT_TRAIN_PARAMS = {
    # default
    "output_dir": None,
    "num_train_epochs": 1,
    "per_device_train_batch_size": 32,
    # "per_device_eval_batch_size": 32,
    # evaluation_strategy: "steps",
    # eval_steps:200,
    "save_steps": 1000,
    "save_total_limit": 2,
    # "load_best_model_at_end": True,
    # metric_for_best_model: "loss",
    # greater_is_better: False,
    "logging_dir": None,
    "logging_steps": 5,
    "gradient_accumulation_steps": 1,
    "learning_rate": 5e-5,
    # disable_tqdm: True,
    # no_cuda: True,
}


class BertTokenizer(TokenizerForHuggingface):
    """
    Examples
    ----------
    >>> tokenizer = BertTokenizer(add_special_tokens=True)
    >>> item = "有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"
    >>> token_item = tokenizer(item)
    >>> print(token_item.input_ids)
    tensor([[ 101, 1062, 2466, 1963, 1745,  138,  100,  140,  166,  117,  167, 5276,
             3338, 3340,  816, 1062, 2466,  102,  168,  134,  166,  116,  128,  167,
             3297, 1920,  966,  138,  100,  140,  102]])
    >>> print(tokenizer.tokenize(item)[:10])
    ['公', '式', '如', '图', '[', '[UNK]', ']', 'x', ',', 'y']
    >>> items = [item, item]
    >>> token_items = tokenizer(items, return_tensors='pt')
    >>> print(token_items.input_ids.shape)
    torch.Size([2, 31])
    >>> print(len(tokenizer.tokenize(items)))
    2
    >>> tokenizer.save_pretrained('test_dir') # doctest: +SKIP
    >>> tokenizer = BertTokenizer.from_pretrained('test_dir') # doctest: +SKIP
    """
    pass


class BertDataset(EduDataset):
    pass


def finetune_bert(items: Union[List[dict], List[str]], output_dir: str, pretrained_model="bert-base-chinese",
                  tokenizer_params=None, data_params=None, model_params=None, train_params=None):
    """
    Parameters
    ----------
    items: list, required
        The training corpus, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_model: str, optional
        The pretrained model name or path for model and tokenizer
    eval_items: list, required
        The evaluating items, each item could be str or dict
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None

    Examples
    ----------
    >>> stems = ["有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$",
    ... "有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$"]
    >>> finetune_bert(stems, "examples/test_model/data/data/bert") # doctest: +SKIP
    {'train_runtime': ..., ..., 'epoch': 1.0}
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    if os.path.exists(pretrained_model):
        tokenizer = BertTokenizer.from_pretrained(pretrained_model, **tokenizer_params)
    else:
        work_tokenizer_params = {
            "add_specials": True,
            "tokenize_method": "pure_text",
        }
        work_tokenizer_params.update(tokenizer_params)
        tokenizer = BertTokenizer(pretrained_model, **work_tokenizer_params)
        # TODO: tokenizer.set_vocab()
    # model configuration
    model = BertForMaskedLM.from_pretrained(pretrained_model, **model_params)
    # resize embedding for additional special tokens
    model.resize_token_embeddings(len(tokenizer.bert_tokenizer))

    # dataset configuration
    dataset = BertDataset(tokenizer, items=items,
                          stem_key=data_params.get("stem_key", None))
    mlm_probability = train_params.pop('mlm_probability', 0.15)
    data_collator = DataCollatorForLanguageModeling(
        tokenizer=tokenizer.bert_tokenizer, mlm=True, mlm_probability=mlm_probability
    )
    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    train_args = TrainingArguments(**work_train_params)
    trainer = Trainer(
        model=model,
        args=train_args,
        data_collator=data_collator,
        train_dataset=dataset,
    )
    trainer.train()
    trainer.save_model(output_dir)
    tokenizer.save_pretrained(output_dir)


def finetune_bert_for_property_prediction(train_items,
                                          output_dir,
                                          pretrained_model="bert-base-chinese",
                                          eval_items=None,
                                          tokenizer_params=None,
                                          data_params=None,
                                          train_params=None,
                                          model_params=None
                                          ):
    """
    Parameters
    ----------
    train_items: list, required
        The training corpus, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_model: str, optional
        The pretrained model name or path for model and tokenizer
    eval_items: list, required
        The evaluating items, each item could be str or dict
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    tokenizer = BertTokenizer.from_pretrained(pretrained_model, **tokenizer_params)
    # dataset configuration
    train_dataset = BertDataset(tokenizer=tokenizer, items=train_items,
                                stem_key=data_params.get("stem_key", "ques_content"),
                                label_key=data_params.get("label_key", "difficulty"))
    if eval_items is not None:
        eval_dataset = BertDataset(tokenizer=tokenizer, items=eval_items,
                                   stem_key=data_params.get("stem_key", "ques_content"),
                                   label_key=data_params.get("label_key", "difficulty"))
    else:
        eval_dataset = None
    # model configuration
    model = BertForPropertyPrediction(pretrained_model, **model_params)
    model.bert.resize_token_embeddings(len(tokenizer.bert_tokenizer))
    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    train_args = TrainingArguments(**work_train_params)
    data_collator = DataCollatorWithPadding(tokenizer.bert_tokenizer)
    trainer = Trainer(
        model=model,
        args=train_args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        data_collator=data_collator,
    )
    trainer.train()
    # trainer.model.save_pretrained(output_dir)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)


def finetune_bert_for_knowledge_prediction(train_items,
                                           output_dir,
                                           pretrained_model="bert-base-chinese",
                                           eval_items=None,
                                           tokenizer_params=None,
                                           data_params=None,
                                           train_params=None,
                                           model_params=None
                                           ):
    """
    Parameters
    ----------
    train_items: list, required
        The training corpus, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_model: str, optional
        The pretrained model name or path for model and tokenizer
    eval_items: list, required
        The evaluating items, each item could be str or dict
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    tokenizer = BertTokenizer.from_pretrained(pretrained_model, **tokenizer_params)
    # dataset configuration
    train_dataset = BertDataset(tokenizer=tokenizer, items=train_items,
                                stem_key=data_params.get("stem_key", "ques_content"),
                                label_key=data_params.get("label_key", "know_list"))
    if eval_items is not None:
        eval_dataset = BertDataset(tokenizer=tokenizer, items=eval_items,
                                   stem_key=data_params.get("stem_key", "ques_content"),
                                   label_key=data_params.get("label_key", "know_list"))
    else:
        eval_dataset = None
    # model configuration
    model = BertForKnowledgePrediction(pretrained_model_dir=pretrained_model, **model_params)
    model.bert.resize_token_embeddings(len(tokenizer.bert_tokenizer))
    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    train_args = TrainingArguments(**work_train_params)
    data_collator = DataCollatorWithPadding(tokenizer.bert_tokenizer)
    trainer = Trainer(
        model=model,
        args=train_args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        data_collator=data_collator,
    )
    trainer.train()
    # trainer.model.save_pretrained(output_dir)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)



from collections import OrderedDict
from ..Pretrain import ElmoTokenizer, BertTokenizer, QuesNetTokenizer, DisenQTokenizer
from ..ModelZoo.rnn import ElmoLMForPropertyPrediction, ElmoLMForKnowledgePrediction
from ..ModelZoo.bert import BertForPropertyPrediction, BertForKnowledgePrediction

TOKENIZER_MAPPING_NAMES = OrderedDict(
    [
        ("elmo", ElmoTokenizer),
        ("bert", BertTokenizer),
        ("quesnet", QuesNetTokenizer),
        ("disenq", DisenQTokenizer)
    ]
)

MODEL_FOR_PROPERTY_PREDICTION_MAPPING_NAMES = OrderedDict(
    [
        ("elmo", ElmoLMForPropertyPrediction),
        ("bert", BertForPropertyPrediction),
    ]
)

MODEL_FOR_KNOWLEDGE_PREDICTION_MAPPING_NAMES = OrderedDict(
    [
        ("elmo", ElmoLMForKnowledgePrediction),
        ("bert", BertForKnowledgePrediction)
    ]
)

TASK_MAPPING = {
    "property-prediction": MODEL_FOR_PROPERTY_PREDICTION_MAPPING_NAMES,
    "knowledge-prediction": MODEL_FOR_KNOWLEDGE_PREDICTION_MAPPING_NAMES
}



import torch
from EduNLP.ModelZoo.disenqnet.disenqnet import DisenQNet


class DisenQModel(object):
    def __init__(self, pretrained_dir, device="cpu"):
        """
        Parameters
        ----------
        pretrained_dir: str
            the dirname to pretrained model
        device: str
            cpu or cuda, default is cpu
        """
        self.device = device
        self.model = DisenQNet.from_pretrained(pretrained_dir)
        self.model.to(self.device)

    def __call__(self, items: dict, **kwargs):
        outputs = self.model(**items)
        return outputs.embeded, outputs.k_hidden, outputs.i_hidden

    def infer_vector(self, items: dict, vector_type=None, **kwargs) -> torch.Tensor:
        """
        Parameters
        ----------
        vector_type: str
            choose the type of items tensor to return.
            Default is None, which means return both (k_hidden, i_hidden)
            when vector_type="k", return k_hidden;
            when vector_type="i", return i_hidden;
        """
        _, k_hidden, i_hidden = self(items)
        if vector_type is None:
            return k_hidden, i_hidden
        elif vector_type == "k":
            return k_hidden
        elif vector_type == "i":
            return i_hidden
        else:
            raise KeyError("vector_type must be one of (None, 'k', 'i') ")

    def infer_tokens(self, items: dict, **kwargs) -> torch.Tensor:
        embeded, _, _ = self(items)
        return embeded

    @property
    def vector_size(self):
        return self.model.hidden_size



import json


def load_items(data_path):
    _data = []
    with open(data_path, encoding="utf-8") as f:
        for line in f.readlines():
            _data.append(json.loads(line))
    return _data



from ..utils import dict2str4sif
from ..SIF import is_sif, to_sif, sif4sci
from ..SIF.segment import seg, SegmentList
from ..Tokenizer import PureTextTokenizer
from ..SIF.tokenization.text import tokenize


class BasePipe:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, input_):
        raise NotImplementedError


class IsSifPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(IsSifPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        print(is_sif(input_, *self.args, **self.kwargs))
        return input_


class ToSifPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(ToSifPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        return to_sif(input_, *self.args, **self.kwargs)


class Dict2Str4SifPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(Dict2Str4SifPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        return dict2str4sif(input_, *self.args, **self.kwargs)


class Sif4SciPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(Sif4SciPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        return sif4sci(input_, *self.args, **self.kwargs)


class SegPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(SegPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        return seg(input_, *self.args, **self.kwargs)


class SegDescribePipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(SegDescribePipe, self).__init__(*args, **kwargs)

    def __call__(self, input_: SegmentList):
        print(input_.describe())
        return input_


class SegFilterPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(SegFilterPipe, self).__init__(*args, **kwargs)

    def __call__(self, input_: SegmentList):
        input_.filter(*self.args, **self.kwargs)
        return input_


class TokenizePipe(BasePipe):
    def __init__(self, *args, **kwargs):
        super(TokenizePipe, self).__init__(*args, **kwargs)

    def __call__(self, input_):
        return tokenize(input_, *self.args, **self.kwargs)


PREPROCESSING_PIPES = {
    'dict2str4sif': Dict2Str4SifPipe,
    'is_sif': IsSifPipe,
    'to_sif': ToSifPipe,
    'sif4sci': Sif4SciPipe,
    'seg': SegPipe,
    'seg_describe': SegDescribePipe,
    'seg_filter': SegFilterPipe,
    'tokenize': TokenizePipe,
}



# coding: utf-8
# 2021/7/12 @ tongshiwei

from .padder import PadSequence, pad_sequence
from .device import set_device
from .masker import Masker
from .data import load_items
from .modules import MLP, TextCNN
from .torch_utils import *



import torch.nn as nn
import json
import os
from pathlib import Path
import torch
from transformers import PretrainedConfig
# import logging
from ..utils import logger


class BaseModel(nn.Module):
    base_model_prefix = ''

    def __init__(self):
        super(BaseModel, self).__init__()
        self.config = PretrainedConfig()

    def forward(self, *input):
        raise NotImplementedError

    def save_pretrained(self, output_dir):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        model_path = os.path.join(output_dir, 'pytorch_model.bin')
        model_path = Path(model_path)
        torch.save(self.state_dict(), model_path.open('wb'))
        self.save_config(output_dir)

    @classmethod
    def from_pretrained(cls, pretrained_model_path, *args, **kwargs):
        config_path = os.path.join(pretrained_model_path, "config.json")
        model_path = os.path.join(pretrained_model_path, "pytorch_model.bin")
        model = cls.from_config(config_path, *args, **kwargs)
        loaded_state_dict = torch.load(model_path)
        loaded_keys = loaded_state_dict.keys()
        expected_keys = model.state_dict().keys()

        prefix = cls.base_model_prefix

        if set(loaded_keys) == set(expected_keys):
            # same architecture
            model.load_state_dict(loaded_state_dict)
        else:
            has_prefix_module = any(s.startswith(prefix) for s in loaded_keys)
            expects_prefix_module = any(s.startswith(prefix) for s in expected_keys)

            new_loaded_state_dict = {}
            if expects_prefix_module and not has_prefix_module:
                # add prefix
                for key in loaded_keys:
                    new_loaded_state_dict['.'.join([prefix, key])] = loaded_state_dict[key]
            if has_prefix_module and not expects_prefix_module:
                # remove prefix
                for key in loaded_keys:
                    if key.startswith(prefix):
                        new_loaded_state_dict['.'.join(key.split('.')[1:])] = loaded_state_dict[key]
            if has_prefix_module and expects_prefix_module:
                # both have prefix, only load the base encoder
                for key in loaded_keys:
                    if key.startswith(prefix):
                        new_loaded_state_dict[key] = loaded_state_dict[key]
            loaded_state_dict = new_loaded_state_dict
            model.load_state_dict(loaded_state_dict, strict=False)
        loaded_keys = loaded_state_dict.keys()
        missing_keys = set(expected_keys) - set(loaded_keys)
        if len(missing_keys) == 0:
            logger.info(
                f"All the weights of {model.__class__.__name__} were initialized from the model checkpoint at"
                f" {pretrained_model_path}.\nIf your task is similar to the task the model of the checkpoint"
                f" was trained on, you can already use {model.__class__.__name__} for predictions without further"
                " training."
            )
        elif len(missing_keys) > 0:
            logger.warning(
                f"Some weights of {model.__class__.__name__} were not initialized from the model checkpoint at"
                f" {pretrained_model_path} and are newly initialized: {missing_keys}\nYou should probably"
                " TRAIN this model on a down-stream task to be able to use it for predictions and inference."
            )
        return model

    def save_config(self, config_dir):
        config_path = os.path.join(config_dir, "config.json")
        with open(config_path, "w", encoding="utf-8") as wf:
            json.dump(self.config.to_dict(), wf, ensure_ascii=False, indent=2)

    @classmethod
    def from_config(cls, config_path, *args, **kwargs):
        raise NotImplementedError



import torch
from torch import nn
from torch.nn import functional as F


class MLP(nn.Module):
    def __init__(self, in_dim, n_classes, hidden_dim, dropout, n_layers=2, act=F.leaky_relu):
        super(MLP, self).__init__()
        self.l_in = nn.Linear(in_dim, hidden_dim)
        self.l_hs = nn.ModuleList(nn.Linear(hidden_dim, hidden_dim) for _ in range(n_layers - 2))  # doctest: +ELLIPSIS
        self.l_out = nn.Linear(hidden_dim, n_classes)
        self.dropout = nn.Dropout(p=dropout)
        self.act = act

    def forward(self, input):
        hidden = self.act(self.l_in(self.dropout(input)))
        for l_h in self.l_hs:
            hidden = self.act(l_h(self.dropout(hidden)))
        output = self.l_out(self.dropout(hidden))
        return output


class TextCNN(nn.Module):
    def __init__(self, embed_dim, hidden_dim):
        super(TextCNN, self).__init__()
        kernel_sizes = [2, 3, 4, 5]
        channel_dim = hidden_dim // len(kernel_sizes)
        self.min_seq_len = max(kernel_sizes)
        self.convs = nn.ModuleList([nn.Conv1d(embed_dim, channel_dim, k_size) for k_size in kernel_sizes])

    def forward(self, embed):
        if embed.size(1) < self.min_seq_len:
            device = embed.device
            pad = torch.zeros(embed.size(0), self.min_seq_len - embed.size(1), embed.size(-1)).to(device)
            embed = torch.cat((embed, pad), dim=1)
        # (b, s, d) => (b, d, s) => (b, d', s') => (b, d', 1) => (b, d')
        # batch_size * dim * seq_len
        hidden = [F.leaky_relu(conv(embed.transpose(1, 2))) for conv in self.convs]
        # batch_size * dim
        hidden = [F.max_pool1d(h, kernel_size=h.size(2)).squeeze(-1) for h in hidden]
        hidden = torch.cat(hidden, dim=-1)
        return hidden



import torch
import os
from copy import deepcopy
from transformers import TrainingArguments, Trainer
from typing import Optional, Union, List
from ..ModelZoo.rnn import ElmoLM, ElmoLMForPreTraining, ElmoLMForPropertyPrediction, ElmoLMForKnowledgePrediction
from ..ModelZoo.utils import pad_sequence
from .pretrian_utils import PretrainedEduTokenizer, EduDataset

__all__ = ["ElmoTokenizer", "ElmoDataset", "train_elmo", "train_elmo_for_property_prediction",
           "train_elmo_for_knowledge_prediction"]

DEFAULT_TRAIN_PARAMS = {
    # default
    "output_dir": None,
    "overwrite_output_dir": True,
    "num_train_epochs": 3,
    "per_device_train_batch_size": 8,
    "per_device_eval_batch_size": 32,
    # evaluation_strategy: "steps",
    # eval_steps:200,
    "save_steps": 1000,
    "save_total_limit": 2,
    # "load_best_model_at_end": False,
    # metric_for_best_model: "loss",
    # greater_is_better: False,
    "logging_dir": None,
    "logging_steps": 5,
    "gradient_accumulation_steps": 1,
    "learning_rate": 5e-4,
    # disable_tqdm: True,
    # no_cuda: True,
}


class ElmoTokenizer(PretrainedEduTokenizer):
    """
    Examples
    --------
    >>> t=ElmoTokenizer()
    >>> items = ["有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"]
    >>> len(t)
    14
    >>> t.tokenize(items[0])
    ['公式', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[SEP]', 'z', '=', 'x', '+', '7', 'y', '最大值', '[MARK]']
    >>> t(items[0])
    {'seq_idx': tensor([1, 1, 6, 1, 1, 1, 1, 1, 9, 1, 1, 1, 1, 1, 1, 1, 7]), 'seq_len': tensor(17)}
    >>> t.set_vocab(items[0])
    ['公式', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[SEP]', 'z', '=', 'x', '+', '7', 'y', '最大值', '[MARK]']
    >>> len(t)
    45
    >>> t(items[0])
    {'seq_idx': tensor([ 1,  1,  6, 26, 27, 28,  1,  1,  9, 35, 36, 26, 37, 38, 28,  1,  7]), 'seq_len': tensor(17)}
    """

    def __init__(self, vocab_path=None, max_length=250, tokenize_method="pure_text", add_specials=True, **kwargs):
        super().__init__(vocab_path=vocab_path, max_length=max_length, tokenize_method=tokenize_method,
                         add_specials=add_specials, **kwargs)


class ElmoDataset(EduDataset):
    def __init__(self, tokenizer: ElmoTokenizer, **kwargs):
        super(ElmoDataset, self).__init__(tokenizer=tokenizer, **kwargs)

    def collate_fn(self, batch_data):
        pad_idx = self.tokenizer.vocab.pad_idx
        first = batch_data[0]
        batch = {
            k: [item[k] for item in batch_data] for k in first.keys()
        }
        batch["seq_idx"] = pad_sequence(batch["seq_idx"], pad_val=pad_idx)
        batch = {key: torch.as_tensor(val) for key, val in batch.items()}
        return batch


def train_elmo(items: Union[List[dict], List[str]], output_dir: str, pretrained_dir: str = None,
               tokenizer_params=None, data_params=None, model_params=None, train_params=None):
    """
    Parameters
    ----------
    items: list, required
        The training corpus, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_dir: str, optional
        The pretrained directory for model and tokenizer
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        - stem_key
        - label_key
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    if pretrained_dir is not None and os.path.exists(pretrained_dir):
        tokenizer = ElmoTokenizer.from_pretrained(pretrained_dir, **tokenizer_params)
    else:
        work_tokenizer_params = {
            "add_specials": True,
            "tokenize_method": "pure_text",
        }
        work_tokenizer_params.update(tokenizer_params if tokenizer_params else {})
        tokenizer = ElmoTokenizer(**work_tokenizer_params)
        corpus_items = items
        if isinstance(items[0], str):
            tokenizer.set_vocab(corpus_items)
        else:
            tokenizer.set_vocab(corpus_items,
                                key=lambda x: x[data_params.get("stem_key", "ques_content")])

    # dataset configuration
    dataset = ElmoDataset(tokenizer=tokenizer, items=items,
                          stem_key=data_params.get("stem_key", "ques_content"))

    # model configuration
    if pretrained_dir:
        model = ElmoLMForPreTraining.from_pretrained(pretrained_dir, **model_params)
    else:
        work_model_params = {
            "vocab_size": len(tokenizer),
            "embedding_dim": 300,
            "hidden_size": 300
        }
        work_model_params.update(model_params if model_params else {})
        model = ElmoLMForPreTraining(**work_model_params)
    model.elmo.LM_layer.rnn.flatten_parameters()

    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    work_args = TrainingArguments(**work_train_params)
    trainer = Trainer(
        model=model,
        args=work_args,
        train_dataset=dataset,
        data_collator=dataset.collate_fn,
    )
    trainer.train()
    # trainer.model.save_pretrained(output_dir)
    assert isinstance(trainer.model, ElmoLMForPreTraining)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)
    return output_dir


def train_elmo_for_property_prediction(
        train_items: list, output_dir: str, pretrained_dir=None, eval_items=None,
        tokenizer_params=None, data_params=None, train_params=None, model_params=None
):
    """
    Parameters
    ----------
    train_items: list, required
        The training items, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_dir: str, optional
        The pretrained directory for model and tokenizer
    eval_items: list, required
        The evaluating items, each item could be str or dict
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    if pretrained_dir is not None:
        tokenizer = ElmoTokenizer.from_pretrained(pretrained_dir, **tokenizer_params)
    else:
        work_tokenizer_params = {
            "add_special_tokens": True,
            "tokenize_method": "pure_text",
        }
        work_tokenizer_params.update(tokenizer_params if tokenizer_params else {})
        tokenizer = ElmoTokenizer(**work_tokenizer_params)
        corpus_items = train_items + eval_items if eval_items else []
        tokenizer.set_vocab(corpus_items,
                            key=lambda x: x[data_params.get("stem_key", "ques_content")])
    # dataset configuration
    train_dataset = ElmoDataset(tokenizer=tokenizer, items=train_items,
                                stem_key=data_params.get("stem_key", "ques_content"),
                                label_key=data_params.get("label_key", "difficulty"))
    if eval_items is not None:
        eval_dataset = ElmoDataset(tokenizer=tokenizer, items=eval_items,
                                   stem_key=data_params.get("stem_key", "ques_content"),
                                   label_key=data_params.get("label_key", "difficulty"))
    else:
        eval_dataset = None
    # model configuration
    if pretrained_dir is not None:
        model = ElmoLMForPropertyPrediction.from_pretrained(pretrained_dir, **model_params)
    else:
        work_model_params = {
            "vocab_size": len(tokenizer),
            "embedding_dim": 512,
            "hidden_size": 512
        }
        work_model_params.update(model_params if model_params else {})
        model = ElmoLMForPropertyPrediction(**work_model_params)
    model.elmo.LM_layer.rnn.flatten_parameters()

    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    work_train_params = TrainingArguments(**work_train_params)
    trainer = Trainer(
        model=model,
        args=work_train_params,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        data_collator=train_dataset.collate_fn,
    )
    trainer.train()
    # trainer.model.save_pretrained(output_dir)
    assert isinstance(trainer.model, ElmoLMForPropertyPrediction)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)
    return output_dir


def train_elmo_for_knowledge_prediction(
        train_items: list, output_dir: str, pretrained_dir=None, eval_items=None,
        tokenizer_params=None, data_params=None, train_params=None, model_params=None
):
    """
    Parameters
    ----------
    train_items: list, required
        The training items, each item could be str or dict
    output_dir: str, required
        The directory to save trained model files
    pretrained_dir: str, optional
        The pretrained directory for model and tokenizer
    eval_items: list, required
        The evaluating items, each item could be str or dict
    tokenizer_params: dict, optional, default=None
        The parameters passed to ElmoTokenizer
    data_params: dict, optional, default=None
        The parameters passed to ElmoDataset and ElmoTokenizer
    model_params: dict, optional, default=None
        The parameters passed to Trainer
    train_params: dict, optional, default=None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    # tokenizer configuration
    if pretrained_dir is not None:
        tokenizer = ElmoTokenizer.from_pretrained(pretrained_dir, **tokenizer_params)
    else:
        work_tokenizer_params = {
            "add_special_tokens": True,
            "tokenize_method": "pure_text",
        }
        work_tokenizer_params.update(tokenizer_params if tokenizer_params else {})
        tokenizer = ElmoTokenizer(**work_tokenizer_params)
        corpus_items = train_items + eval_items if eval_items else []
        tokenizer.set_vocab(corpus_items,
                            key=lambda x: x[data_params.get("stem_key", "ques_content")])
    # dataset configuration
    train_dataset = ElmoDataset(tokenizer=tokenizer, items=train_items,
                                stem_key=data_params.get("stem_key", "ques_content"),
                                label_key=data_params.get("label_key", "know_list"))
    if eval_items is not None:
        eval_dataset = ElmoDataset(tokenizer=tokenizer, items=eval_items,
                                   stem_key=data_params.get("stem_key", "ques_content"),
                                   label_key=data_params.get("label_key", "know_list"))
    else:
        eval_dataset = None
    # model configuration
    if pretrained_dir is not None:
        model = ElmoLMForKnowledgePrediction.from_pretrained(pretrained_dir, **model_params)
    else:
        work_model_params = {
            "vocab_size": len(tokenizer),
            "embedding_dim": 512,
            "hidden_size": 512
        }
        work_model_params.update(model_params if model_params else {})
        model = ElmoLMForKnowledgePrediction(**work_model_params)
    model.elmo.LM_layer.rnn.flatten_parameters()

    # training configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    work_train_params = TrainingArguments(**work_train_params)
    trainer = Trainer(
        model=model,
        args=work_train_params,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        data_collator=train_dataset.collate_fn,
    )
    trainer.train()
    assert isinstance(trainer.model, ElmoLMForKnowledgePrediction)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)
    return output_dir



# coding: utf-8
# 2021/8/2 @ tongshiwei
import pytest
from EduNLP import get_pretrained_i2v
# from EduNLP.I2V.i2v import MODELS
from EduNLP.I2V import D2V, W2V
from EduNLP.Vector import get_pretrained_model_info, get_all_pretrained_models


def test_pretrained_i2v(tmp_path):

    d = tmp_path / "model"
    d.mkdir()

    url, t2v_name = get_pretrained_model_info("d2v_test_256")
    assert url != ""
    assert t2v_name == "d2v"
    model_names = get_all_pretrained_models()
    assert "d2v_test_256" in model_names

    get_pretrained_i2v("d2v_test_256", d)

    with pytest.raises(KeyError):
        get_pretrained_i2v("error")

    get_pretrained_i2v("w2v_test_256", d)

#     get_pretrained_i2v("quesnet_test_256", d)

#     get_pretrained_i2v("elmo_test", d)

#     # get_pretrained_i2v("tal_edu_bert", d)



# coding: utf-8
# 2021/5/29 @ tongshiwei

from .gensim_vec import train_vector, GensimWordTokenizer, GensimSegTokenizer
from .elmo_vec import *
from .bert_vec import *
from .quesnet_vec import QuesNetTokenizer, pretrain_quesnet, Question
from .disenqnet_vec import DisenQTokenizer, train_disenqnet
from .pretrian_utils import *
from .hugginface_utils import *
